VERSION = '1.1.1'
import os
import sys
import logging
import argparse as ap

os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2' 

from Cellist.version import __version__
from Cellist.Segmentation import CellistParser, Cellist
from Cellist.Registration import AlignmentParser, Alignment
from Cellist.Watershed import WatershedParser, Watershed
from Cellist.Imputation import ImputationParser, Imputation
from Cellist.Cellpose import CellposeParser, Cellpose


def main():
    """
    Add main function argument parsers.
    """
    parser = ap.ArgumentParser(description = "Cellist (Cell identification in high-resolution Spatial Transcriptomics) is a cell segmentation tool for high-resolution spatial transcriptomics. ")
    parser.add_argument("-v", "--version", action = "store_true", help = "Print version info.")

    subparsers = parser.add_subparsers(dest = "subcommand")
    CellistParser(subparsers)
    AlignmentParser(subparsers)
    WatershedParser(subparsers)
    CellposeParser(subparsers)
    ImputationParser(subparsers)

    logging.basicConfig(format="%(levelname)s: %(message)s", stream=sys.stderr)
    args = parser.parse_args()

    version = __version__

    if args.version:
        print(version)
        exit(0)
    elif args.subcommand == "seg":
        platform = args.platform
        resolution = args.resolution
        spot_expr_file = args.spot_expr_file
        all_spot_count_h5_file = args.all_spot_count_h5_file
        nucleus_seg_method = args.nucleus_seg_method
        props_file = args.props_file
        nucleus_count_h5_file = args.nucleus_count_h5_file
        nucleus_coord_file = args.nucleus_coord_file
        patch_data_dir = args.patch_data_dir
        num_workers = args.num_workers
        max_dist = args.max_dist
        neigh_dist = args.neigh_dist
        noise_prop = args.noise_prop
        alpha = args.alpha
        sigma = args.sigma
        beta = args.beta
        gene_use = args.gene_use
        two_step = args.two_step
        cyto = args.cyto
        out_dir = args.out_dir
        out_prefix = args.out_prefix
        max_dist_s1_scale = args.max_dist_s1_scale
        noise_prop_s1 = args.noise_prop_s1
        Cellist(platform = platform, resolution = resolution, nucleus_seg_method = nucleus_seg_method, 
            props_file = props_file, nucleus_count_h5_file = nucleus_count_h5_file, 
            nucleus_coord_file = nucleus_coord_file, all_spot_count_h5_file = all_spot_count_h5_file, 
            spot_expr_file = spot_expr_file, patch_data_dir = patch_data_dir, num_workers = num_workers, 
            alpha = alpha, sigma = sigma, beta = beta, 
            gene_use = gene_use, max_dist = max_dist, 
            two_step = two_step, cyto = cyto, max_dist_s1_scale = max_dist_s1_scale, 
            noise_prop_s1 = noise_prop_s1, noise_prop = noise_prop, neigh_dist = neigh_dist,
            out_dir = out_dir, out_prefix = out_prefix)
    elif args.subcommand == "align":
        gem_path = args.spot_expr_file
        img_path = args.regist_tif_file
        num_workers = args.num_workers
        out_dir = args.out_dir
        out_prefix = args.out_prefix
        Alignment(gem_path = gem_path, img_path = img_path, num_workers = num_workers, 
            out_dir = out_dir, out_prefix = out_prefix)
    elif args.subcommand == "watershed":
        platform = args.platform
        gem_path = args.spot_expr_file
        img_path = args.regist_tif_file
        out_dir = args.out_dir
        out_prefix = args.out_prefix
        min_distance = args.min_distance
        no_local_threshold = args.no_local_threshold
        expansion = args.expansion
        expansion_dist = args.expansion_dist
        Watershed(platform = platform, gem_path = gem_path, img_path = img_path, out_dir = out_dir, out_prefix = out_prefix,
        min_distance = min_distance, no_local_threshold = no_local_threshold, expansion = expansion, expansion_dist = expansion_dist)
    elif args.subcommand == "cellpose":
        platform = args.platform
        gem_path = args.spot_expr_file
        img_path = args.regist_tif_file
        out_dir = args.out_dir
        out_prefix = args.out_prefix
        no_local_threshold = args.no_local_threshold
        diameter = args.diameter
        model_type = args.model_type
        flow_threshold = args.flow_threshold
        cellprob_threshold = args.cellprob_threshold
        expansion = args.expansion
        expansion_dist = args.expansion_dist
        Cellpose(platform = platform, gem_path = gem_path, img_path = img_path, out_dir = out_dir, out_prefix = out_prefix, 
        no_local_threshold = no_local_threshold, diameter = diameter, model_type = model_type, flow_threshold = flow_threshold, cellprob_threshold = cellprob_threshold,
        expansion = expansion, expansion_dist = expansion_dist)
    elif args.subcommand == "impute":
        expr_file = args.expr
        spatial_file = args.spatial
        num_workers = args.num_workers
        out_dir = args.out_dir
        out_prefix = args.out_prefix
        Imputation(expr_file = expr_file, spatial_file = spatial_file, num_workers = num_workers, 
            out_dir = out_dir, out_prefix = out_prefix)
    else:
        parser.print_help()
        exit(1)
    exit(0)

if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
# @Author: dongqing
# @Date:   2023-10-14 11:57:10
# @Last Modified by:   Dongqing
# @Last Modified time: 2025-08-26 13:58:25


import pandas as pd
import numpy as np
from skimage import color
import matplotlib
import matplotlib.pyplot as plt
import os
import h5py
from mpl_toolkits.axes_grid1.axes_divider import make_axes_locatable

matplotlib.rcParams['font.family'] = 'sans-serif'
matplotlib.rcParams['font.sans-serif'] = ['Arial']
matplotlib.rcParams["legend.frameon"] = False
matplotlib.rcParams['pdf.fonttype'] = 42

def hex_to_rgb(hex_str):
    rgb = []
    for i in (1, 3, 5):
        decimal = int(hex_str[i:i+2], 16)/255
        rgb.append(decimal)
    return tuple(rgb)

zeileis_28 = [
    "#023fa5", "#7d87b9", "#bec1d4", "#d6bcc0", 
    "#bb7784", "#8e063b", "#4a6fe3", "#8595e1", 
    "#b5bbe3", "#e6afb9", "#e07b91", "#d33f6a", 
    "#11c638", "#8dd593", "#c6dec7", "#ead3c6", 
    "#f0b98d", "#ef9708", "#0fcfc0", "#9cded6", 
    "#d5eae7", "#f3e1eb", "#f6c4e1", "#f79cd4", 
    '#7f7f7f', "#c7c7c7", "#1CE6FF", "#336600",
]
zeileis_28_rgb = [hex_to_rgb(i) for i in zeileis_28]


def draw_segmentation(coord_df_sub, seg_res, out_prefix, out_dir, x = "X_img", y = "Y_img", figsize = (15, 15)):
    seg_color = zeileis_28_rgb
    community_seg = coord_df_sub.pivot(index = x, columns = y, values = seg_res)
    row_to_add = list(set(range(coord_df_sub[x].min(), coord_df_sub[x].max() + 1)) - set(community_seg.index))
    community_seg_row_to_add = pd.DataFrame(np.zeros((len(row_to_add), community_seg.shape[1])), index = row_to_add, columns = community_seg.columns)
    community_seg = pd.concat([community_seg, community_seg_row_to_add])
    community_seg = community_seg.loc[sorted(community_seg.index.tolist()), :]
    col_to_add = list(set(range(coord_df_sub[y].min(), coord_df_sub[y].max() + 1))  - set(community_seg.columns))
    community_seg.loc[:, col_to_add] = np.nan
    community_seg = community_seg.loc[:, sorted(community_seg.columns.tolist())]
    community_seg = community_seg.fillna(0)
    fig, ax = plt.subplots(1, 1, figsize=figsize)
    ax.imshow(color.label2rgb(np.array(community_seg), colors = seg_color, bg_label=0))
    ax.axis('off')
    plt.tight_layout()
    plt.savefig(os.path.join(out_dir, "%s_segmentation_plot.pdf" %out_prefix))
    plt.savefig(os.path.join(out_dir, "%s_segmentation_plot.png" %out_prefix))

def draw_segmentation_prob(coord_df_sub, col_prob, out_prefix, out_dir, x = "X_img", y = "Y_img", figsize = (16, 15)):
    community_seg = coord_df_sub.pivot(index = x, columns = y, values = col_prob)
    row_to_add = list(set(range(coord_df_sub[x].min(), coord_df_sub[x].max() + 1)) - set(community_seg.index))
    community_seg_row_to_add = pd.DataFrame(np.zeros((len(row_to_add), community_seg.shape[1])), index = row_to_add, columns = community_seg.columns)
    community_seg = pd.concat([community_seg, community_seg_row_to_add])
    community_seg = community_seg.loc[sorted(community_seg.index.tolist()), :]
    col_to_add = list(set(range(coord_df_sub[y].min(), coord_df_sub[y].max() + 1))  - set(community_seg.columns))
    community_seg.loc[:, col_to_add] = np.nan
    community_seg = community_seg.loc[:, sorted(community_seg.columns.tolist())]
    community_seg = community_seg.fillna(0)
    fig, ax = plt.subplots(1, 1, figsize=figsize)
    im = ax.imshow(np.array(community_seg))
    divider = make_axes_locatable(ax)
    cax = divider.append_axes("right", size="5%", pad=0.05)
    ax.axis('off')
    plt.colorbar(im, cax = cax)
    plt.tight_layout()
    plt.savefig(os.path.join(out_dir, "%s_segmentation_probability_plot.pdf" %out_prefix))


def draw_segmentation_boundary(img_sub, coord_df_sub, seg_res, out_prefix, out_dir, x = "X_img", y = "Y_img", mode = "outer", figsize = (15, 15)):
    seg_color = zeileis_28_rgb
    community_seg = coord_df_sub.pivot(index = x, columns = y, values = seg_res)
    row_to_add = list(set(range(coord_df_sub[x].min(), coord_df_sub[x].max() + 1)) - set(community_seg.index))
    community_seg_row_to_add = community_seg.iloc[range(len(row_to_add)), :]
    community_seg_row_to_add.index = row_to_add
    community_seg_row_to_add.loc[:, :] = 0
    community_seg = pd.concat([community_seg, community_seg_row_to_add], ignore_index=True)
    community_seg = community_seg.loc[sorted(community_seg.index.tolist()), :]
    col_to_add = list(set(range(coord_df_sub[y].min(), coord_df_sub[y].max() + 1))  - set(community_seg.columns))
    community_seg.loc[:, col_to_add] = np.nan
    community_seg = community_seg.loc[:, sorted(community_seg.columns.tolist())]
    community_seg = community_seg.fillna(0)
    fig, ax = plt.subplots(1, 1, figsize=figsize)
    ax.imshow(segmentation.mark_boundaries(img_sub, np.array(community_seg).astype(int), mode = mode ,color = (1,0,0), background = 0))
    ax.axis('off')
    plt.tight_layout()
    plt.savefig(os.path.join(out_dir, "%s_segmentation_boundary_plot.pdf" %out_prefix))



# -*- coding: utf-8 -*-
# @Author: dongqing
# @Date:   2023-08-26 22:39:03
# @Last Modified by:   dongqing
# @Last Modified time: 2025-04-28 20:02:19

import os
import matplotlib
import pandas as pd
import numpy as np
import scanpy as sc
import anndata as ad
import matplotlib.pyplot as plt

from scipy.spatial import KDTree
from scipy import sparse
import concurrent.futures


def sub_mat(mat, genes, cells, genes_sub = None, cells_sub = None):
	mat_use = mat
	if not genes_sub:
		if type(genes[0]) == bytes:
			genes = [i.decode() for i in genes]
		genes_sub = genes	
	if not cells_sub:
		if type(genes[0]) == bytes:
			cells = [i.decode() for i in cells]
		cells_sub = cells
	genes_used = sorted(genes_sub)
	genes_used_array = np.array(genes_used)
	genes_array = np.array(genes)
	genes_sorter = np.argsort(genes_array)
	genes_used_index = genes_sorter[np.searchsorted(genes_array, genes_used_array, sorter = genes_sorter)]
	mat_use = mat_use[genes_used_index, :]
	cells_used = sorted(cells_sub)
	cells_used_array = np.array(cells_used)
	cells_array = np.array(cells)
	cells_sorter = np.argsort(cells_array)
	cells_used_index = cells_sorter[np.searchsorted(cells_array, cells_used_array, sorter = cells_sorter)]
	mat_use = mat_use[:, cells_used_index]
	return(mat_use, genes_used, cells_used)

def longdf_to_mat(gene_cell_df):
	cells = set(str(x[0]) for x in gene_cell_df.index)
	genes = set(str(x[1]) for x in gene_cell_df.index)
	cells_dict = dict(zip(cells, range(0, len(cells))))
	genes_dict = dict(zip(genes, range(0, len(genes))))
	rows = [cells_dict[str(x[0])] for x in gene_cell_df.index]
	cols = [genes_dict[str(x[1])] for x in gene_cell_df.index]
	expr_mat = sparse.csr_matrix((gene_cell_df.values, (rows, cols)))
	expr_mat= expr_mat.T
	gene_list = list(genes)
	cell_list = list(cells)
	return(expr_mat, gene_list, cell_list)

def SpotBinGene(gem_df, bin_size, countname = 'MIDCount'):
	gem_df['x_bin'] = (gem_df['x']/bin_size).astype(np.uint32)*bin_size
	gem_df['y_bin'] = (gem_df['y']/bin_size).astype(np.uint32)*bin_size
	gem_df['binID'] = gem_df['x_bin'].astype(str) + "_" + gem_df['y_bin'].astype(str)
	gem_df_bin_gene = gem_df[countname].groupby([gem_df['binID'], gem_df['geneID']]).sum()
	expr_mat, gene_list, cell_list = longdf_to_mat(gem_df_bin_gene)
	return(gem_df, expr_mat, gene_list, cell_list)

def cal_dist(coord_df, mdist):
	kd_tree = KDTree(coord_df)
	sdm = kd_tree.sparse_distance_matrix(kd_tree, mdist)
	sdm_csc = sparse.csc_matrix(sdm, dtype=np.float32)
	return(kd_tree, sdm_csc)

def cal_weight(sdm_csc):
	weight = sdm_csc.copy()
	np.reciprocal(weight.data, out=weight.data)
	np.power(weight.data, 2, out=weight.data)
	weight.setdiag(1)
	return(weight)

def enhance_spot_expr(coord_df, coord_expr_mat, neigh_dist = 5, x_colname = "x", y_colname = "y"):
	kd_tree, sdm_csc = cal_dist(coord_df[[x_colname, y_colname]], neigh_dist)
	weight = cal_weight(sdm_csc)
	coord_expr_mat_enhanced = coord_expr_mat.dot(weight)
	coord_expr_mat_enhanced_array = coord_expr_mat_enhanced.toarray()
	return(coord_expr_mat_enhanced_array)

def pearson_corr(x_array, y_array):
	xmean = x_array.mean(axis = 0)
	ymean = y_array.mean(axis = 0)
	xm = x_array - xmean
	ym = y_array - ymean
	normxm = np.linalg.norm(xm, axis = 0)
	normym = np.linalg.norm(ym, axis = 0)
	x_normalized = xm/normxm
	y_normalized = ym/normym
	r = x_normalized.T @ y_normalized
	return(r)

def pearson_corr_sparse(x_array, y_array):
	xmean = x_array.mean(axis = 0)
	ymean = y_array.mean(axis = 0)
	xm = x_array - xmean
	ym = y_array - ymean
	xm = sparse.csc_matrix(xm)
	ym = sparse.csc_matrix(ym)
	normxm = sparse.linalg.norm(xm, axis = 0)
	normym = sparse.linalg.norm(ym, axis = 0)
	r = np.dot((xm/normxm).T, ym/normym)
	return(r)

def slice_coord(coord_df, span_length, overlap_length, x_colname = "x", y_colname = "y"):
	xmin = coord_df[x_colname].min()
	xmax = coord_df[x_colname].max()
	ymin = coord_df[y_colname].min()
	ymax = coord_df[y_colname].max()
	i = xmin
	coord_sliced_list = []
	while i + span_length <= xmax:
		xmin_sliced = i
		xmax_sliced = i + span_length
		j = ymin
		while j + span_length <= ymax:
			ymin_sliced = j
			ymax_sliced = j + span_length
			coord_sliced_list.append((xmin_sliced, xmax_sliced, ymin_sliced, ymax_sliced))
			j = j + span_length - overlap_length
		ymin_sliced = j
		ymax_sliced = ymax
		coord_sliced_list.append((xmin_sliced, xmax_sliced, ymin_sliced, ymax_sliced))
		i = i + span_length - overlap_length
	xmin_sliced = i
	xmax_sliced = xmax + 1
	j = ymin
	while j + span_length <= ymax:
		ymin_sliced = j
		ymax_sliced = j + span_length
		coord_sliced_list.append((xmin_sliced, xmax_sliced, ymin_sliced, ymax_sliced))
		j = j + span_length - overlap_length
	ymin_sliced = j
	ymax_sliced = ymax + 1
	coord_sliced_list.append((xmin_sliced, xmax_sliced, ymin_sliced, ymax_sliced))
	return(coord_sliced_list)

def filter_mat(all_count_mat, all_count_genes, all_count_spots, sub_spots = None, gene_pct = 0.05):
	if sub_spots:
		sub_expr, sub_gene, sub_spot = sub_mat(mat = all_count_mat, 
											   genes = all_count_genes, 
											   cells = all_count_spots, 
											   cells_sub = sub_spots)
	else:
		sub_expr, sub_gene, sub_spot = all_count_mat, all_count_genes, all_count_spots
	sub_gene_inspot = sub_expr.sum(axis = 1)
	nSpot = np.quantile(np.array(sub_gene_inspot.flatten())[0], (1 - gene_pct))
	sub_gene_over = np.array(sub_gene)[np.array((sub_gene_inspot > nSpot).flatten())[0]]
	sub_expr_over, sub_gene_over_list, sub_spot_over = sub_mat(mat = sub_expr, 
															   genes = sub_gene, 
															   cells = sub_spot, 
															   genes_sub = sub_gene_over.tolist())
	sub_spot_ngene = sub_expr.sum(axis = 0)
	sub_spot_over0 = np.array(sub_spot_over)[np.array((sub_spot_ngene > 0).flatten())[0]]
	sub_expr_over0, sub_gene_over_list0, sub_spot_over0_list = sub_mat(mat = sub_expr_over, 
															   genes = sub_gene_over_list, 
															   cells = sub_spot_over, 
															   cells_sub = sub_spot_over0.tolist())
	return(sub_expr_over0, sub_gene_over_list0, sub_spot_over0_list)

def get_hvg(all_count_mat, all_count_genes, all_count_spots, n_top_genes):
	all_count_ad = ad.AnnData(
		all_count_mat.T,
		obs=pd.DataFrame(index=all_count_spots),
		var=pd.DataFrame(index=all_count_genes),
	)
	# sc.pp.filter_cells(all_count_ad, min_genes=200)
	# sc.pp.filter_genes(all_count_ad, min_cells=3)
	# sc.pp.normalize_total(all_count_ad, target_sum=1e4)
	# sc.pp.log1p(all_count_ad)
	sc.pp.highly_variable_genes(all_count_ad, n_top_genes=n_top_genes, flavor='seurat_v3', span=1.0)
	hvg = all_count_ad.var[all_count_ad.var.highly_variable].index.tolist()
	return(hvg)

def get_frequent_gene(all_count_mat, all_count_genes, all_count_spots, gene_pct):
	all_count_mat[all_count_mat > 0] = 1
	gene_inspot = all_count_mat.sum(axis = 1)
	nSpot = np.quantile(np.array(gene_inspot.flatten())[0], (1 - gene_pct))
	freq_gene = np.array(all_count_genes)[np.array((gene_inspot > nSpot).flatten())[0]]
	return(freq_gene.tolist())

def get_cell_dist(props_df_sub):
	cell_cell_dist = cal_dist(props_df_sub.loc[:,('centroid-0','centroid-1')], 100)[1]
	cell_cell_dist_array = cell_cell_dist.toarray()
	props_df_sub.loc[:, 'min_cell_dist'] = np.amin(cell_cell_dist_array, axis = 1, where = cell_cell_dist_array!=0, initial = 100)
	props_df_sub.loc[:, 'min_cell_dist'] = np.round(props_df_sub['min_cell_dist'], 2)
	cell_cell_dist_median = np.median(props_df_sub['min_cell_dist'])
	max_dist = cell_cell_dist_median/2
	return(max_dist)

def filter_mat_by_var(all_count_mat, all_count_genes, all_count_spots, sub_spots = None):
	if sub_spots:
		sub_expr, sub_gene, sub_spot = sub_mat(mat = all_count_mat, 
											   genes = all_count_genes, 
											   cells = all_count_spots, 
											   cells_sub = sub_spots)
	else:
		sub_expr, sub_gene, sub_spot = all_count_mat, all_count_genes, all_count_spots
	all_count_ad = ad.AnnData(
		all_count_mat.T,
		obs=pd.DataFrame(index=all_count_spots),
		var=pd.DataFrame(index=all_count_genes),
	)
	sc.pp.highly_variable_genes(all_count_ad, n_top_genes=2000, flavor='seurat_v3', span=1.0)
	selected_gene = all_count_ad.var[all_count_ad.var.highly_variable].index
	sub_expr_over, sub_gene_over_list, sub_spot_over = sub_mat(mat = sub_expr, 
															   genes = sub_gene, 
															   cells = sub_spot, 
															   genes_sub = selected_gene.tolist())
	sub_spot_ngene = sub_expr_over.sum(axis = 0)
	sub_spot_over0 = np.array(sub_spot_over)[np.array((sub_spot_ngene > 0).flatten())[0]]
	sub_expr_over0, sub_gene_over_list0, sub_spot_over0_list = sub_mat(mat = sub_expr_over, 
															   genes = sub_gene_over_list, 
															   cells = sub_spot_over, 
															   cells_sub = sub_spot_over0.tolist())
	return(sub_expr_over0, sub_gene_over_list0, sub_spot_over0_list)

def cal_corr_within_seg(count_df_sub, seg_res, count_name = 'MIDCount', genes_list = None):
	count_df_sub_in = count_df_sub[count_df_sub[seg_res].notna()]
	count_df_sub_in_1 = count_df_sub[count_df_sub['Random_divide'] == 1]
	count_df_sub_in_2 = count_df_sub[count_df_sub['Random_divide'] == 2]
	gene_seg_cell_1 = count_df_sub_in_1[count_name].groupby([count_df_sub_in_1[seg_res], count_df_sub_in_1['geneID']]).sum()
	seg_count_mat_1, seg_count_genes_1, seg_count_cells_1 = longdf_to_mat(gene_seg_cell_1)
	gene_seg_cell_2 = count_df_sub_in_2[count_name].groupby([count_df_sub_in_2[seg_res], count_df_sub_in_2['geneID']]).sum()
	seg_count_mat_2, seg_count_genes_2, seg_count_cells_2 = longdf_to_mat(gene_seg_cell_2)
	if genes_list:
		genes_overlap = list((set(genes_list) & set(seg_count_genes_1)) & set(seg_count_genes_2))
	else:
		genes_overlap = list(set(seg_count_genes_1) & set(seg_count_genes_2))
	cells_overlap = list(set(seg_count_cells_1) & set(seg_count_cells_2))
	seg_count_mat_1_overlap, seg_count_gene_1_overlap, seg_count_cell_1_overlap = sub_mat(mat = seg_count_mat_1, 
								   genes = seg_count_genes_1, 
								   cells = seg_count_cells_1, 
								   cells_sub = cells_overlap, 
								   genes_sub = genes_overlap)
	seg_count_mat_2_overlap, seg_count_gene_2_overlap, seg_count_cell_2_overlap = sub_mat(mat = seg_count_mat_2, 
								   genes = seg_count_genes_2, 
								   cells = seg_count_cells_2, 
								   cells_sub = cells_overlap, 
								   genes_sub = genes_overlap)
	corr_random = pearson_corr(seg_count_mat_1_overlap.toarray(), seg_count_mat_2_overlap.toarray()).diagonal()
	return(corr_random, seg_count_cell_1_overlap)

def cal_corr_within_seg_nucleus(count_df_sub, seg_res, genes_list, count_name = 'MIDCount'):
	count_df_sub_in = count_df_sub[count_df_sub[seg_res].notna()]
	count_df_sub_in_1 = count_df_sub_in[count_df_sub_in['Nucleus'] == 0]
	count_df_sub_in_2 = count_df_sub_in[count_df_sub_in['Nucleus'] == 1]
	gene_seg_cell_1 = count_df_sub_in_1[count_name].groupby([count_df_sub_in_1[seg_res], count_df_sub_in_1['geneID']]).sum()
	seg_count_mat_1, seg_count_genes_1, seg_count_cells_1 = longdf_to_mat(gene_seg_cell_1)
	gene_seg_cell_2 = count_df_sub_in_2[count_name].groupby([count_df_sub_in_2[seg_res], count_df_sub_in_2['geneID']]).sum()
	seg_count_mat_2, seg_count_genes_2, seg_count_cells_2 = longdf_to_mat(gene_seg_cell_2)
	genes_overlap = list((set(genes_list) & set(seg_count_genes_1)) & set(seg_count_genes_2))
	cells_overlap = list(set(seg_count_cells_1) & set(seg_count_cells_2))
	seg_count_mat_1_overlap, seg_count_gene_1_overlap, seg_count_cell_1_overlap = sub_mat(mat = seg_count_mat_1, 
	                               genes = seg_count_genes_1, 
	                               cells = seg_count_cells_1, 
	                               cells_sub = cells_overlap, 
	                               genes_sub = genes_overlap)
	seg_count_mat_2_overlap, seg_count_gene_2_overlap, seg_count_cell_2_overlap = sub_mat(mat = seg_count_mat_2, 
	                               genes = seg_count_genes_2, 
	                               cells = seg_count_cells_2, 
	                               cells_sub = cells_overlap, 
	                               genes_sub = genes_overlap)
	corr_random = pearson_corr(seg_count_mat_1_overlap, seg_count_mat_2_overlap).A.diagonal()
	return((corr_random, seg_count_cell_2_overlap))

def cal_corr_between_seg(count_df_sub, seg_res_1, seg_res_2, count_name = 'MIDCount', genes_list = None):
	count_df_sub_overlap = count_df_sub.loc[count_df_sub[seg_res_1] == count_df_sub[seg_res_2], :]
	count_df_sub_nonoverlap = count_df_sub.loc[count_df_sub[seg_res_1] != count_df_sub[seg_res_2], :]
	# overlap
	count_df_sub_overlap = count_df_sub_overlap[count_df_sub_overlap[seg_res_1].notna()]
	gene_seg_cell_overlap = count_df_sub_overlap[count_name].groupby([count_df_sub_overlap[seg_res_1], count_df_sub_overlap['geneID']]).sum()
	seg_count_mat_overlap, seg_count_genes_overlap, seg_count_cells_overlap = longdf_to_mat(gene_seg_cell_overlap)
	# segmentation 1
	count_df_sub_nonoverlap_1 = count_df_sub_nonoverlap[count_df_sub_nonoverlap[seg_res_1].notna()]
	gene_seg_cell_1 = count_df_sub_nonoverlap_1[count_name].groupby([count_df_sub_nonoverlap_1[seg_res_1], count_df_sub_nonoverlap_1['geneID']]).sum()
	seg_count_mat_1, seg_count_genes_1, seg_count_cells_1 = longdf_to_mat(gene_seg_cell_1)
	# segmentation 2
	count_df_sub_nonoverlap_2 = count_df_sub_nonoverlap[count_df_sub_nonoverlap[seg_res_2].notna()]
	gene_seg_cell_2 = count_df_sub_nonoverlap_2[count_name].groupby([count_df_sub_nonoverlap_2[seg_res_2], count_df_sub_nonoverlap_2['geneID']]).sum()
	seg_count_mat_2, seg_count_genes_2, seg_count_cells_2 = longdf_to_mat(gene_seg_cell_2)
	# determine cells and genes to use
	cells_overlap = sorted(np.intersect1d(np.intersect1d(seg_count_cells_overlap, seg_count_cells_1), seg_count_cells_2).tolist())
	genes_overlap = sorted(np.intersect1d(np.intersect1d(seg_count_genes_overlap, seg_count_genes_1), seg_count_genes_2).tolist())
	if genes_list:
		genes_use = sorted(np.intersect1d(genes_overlap, np.array(genes_list)).tolist())
	else:
		genes_use = genes_overlap
	seg_count_mat_overlap_use, seg_count_genes_overlap_use, seg_count_cells_overlap_use = sub_mat(
											  mat = seg_count_mat_overlap, 
											  genes = seg_count_genes_overlap, 
											  cells = seg_count_cells_overlap, 
											  cells_sub = cells_overlap, 
											  genes_sub = genes_use)
	seg_count_mat_1_use, seg_count_genes_1_use, seg_count_cells_1_use = sub_mat(
											  mat = seg_count_mat_1, 
											  genes = seg_count_genes_1, 
											  cells = seg_count_cells_1, 
											  cells_sub = cells_overlap, 
											  genes_sub = genes_use)
	seg_count_mat_2_use, seg_count_genes_2_use, seg_count_cells_2_use = sub_mat(
											  mat = seg_count_mat_2, 
											  genes = seg_count_genes_2, 
											  cells = seg_count_cells_2, 
											  cells_sub = cells_overlap, 
											  genes_sub = genes_use)
	corr_1 = pearson_corr(seg_count_mat_overlap_use, seg_count_mat_1_use).A.diagonal()
	corr_2 = pearson_corr(seg_count_mat_overlap_use, seg_count_mat_2_use).A.diagonal()
	return((corr_1, corr_2, seg_count_cells_2_use))

def KL_divergence(X, Y):
	X = X + 0.01
	Y = Y + 0.01
	X = X/X.sum(axis=1, keepdims=True)
	Y = Y/Y.sum(axis=1, keepdims=True)
	log_X = np.log(X)
	log_Y = np.log(Y)
	X_log_X = np.matrix((X*log_X).sum(axis = 1))
	D = X_log_X.T - np.dot(X,log_Y.T)
	return np.asarray(D)

def KL_divergence_rm0(X, Y):
	d_list = []
	for x_idx in range(X.shape[0]):
		d_x_list = []
		for y_idx in range(Y.shape[0]):
			x = X[x_idx, :]
			y = Y[y_idx, :]
			gene_union = np.union1d(np.where(x !=0), np.where(y !=0))
			x = x[gene_union]
			y = y[gene_union]
			x = x + 0.01
			y = y + 0.01
			x = x/x.sum()
			y = y/y.sum()
			log_x = np.log(x)
			log_y = np.log(y)
			x_log_x = (x*log_x).sum()
			d = x_log_x.T - np.dot(x,log_y.T)
			d_x_list.append(d)
		d_list.append(d_x_list)
	D = np.array(d_list)
	return(D)

def get_cell_mat(count_df_seg, seg_res, count_name = "MIDCount"):
	count_df_seg = count_df_seg.dropna(subset = [seg_res])
	gene_cell = count_df_seg[count_name].groupby([count_df_seg[seg_res], count_df_seg['geneID']]).sum()
	expr_mat, gene_list, cell_list = longdf_to_mat(gene_cell)
	return(expr_mat, gene_list, cell_list)

def split_cell_with_line(coord_df_sub_cell, seg_res):
    """
    Splits a cell by a randomly oriented line and assigns labels based on projections.
    
    Parameters:
    - coord_df_sub_cell (pd.DataFrame): Subset DataFrame containing coordinates of a single cell.
    - seg_res (str): Column name representing the segmentation result.
    
    Returns:
    - pd.DataFrame: Modified DataFrame with a new 'Random_{seg_res}' column.
    """
    # Calculate the center point
    cx, cy = coord_df_sub_cell[['x', 'y']].mean()
    
    # Generate a random angle for the line
    line_angle = np.random.uniform(0, np.pi)
    line_normal = np.array([np.cos(line_angle), np.sin(line_angle)])
    
    # Calculate projections of points onto the line
    points_centered = coord_df_sub_cell[['x', 'y']].to_numpy() - np.array([cx, cy])
    projections = points_centered @ line_normal  # Using matrix multiplication for efficiency
    
    # Assign labels based on projection values
    coord_df_sub_cell = coord_df_sub_cell.copy()
    coord_df_sub_cell[f'Random_{seg_res}'] = np.where(projections >= 0, 2, 1)
    
    return coord_df_sub_cell

def split_cells_with_line(coord_df, seg_res, num_workers=16):
    """
    Splits multiple cells by randomly oriented lines in parallel.
    
    Parameters:
    - coord_df (pd.DataFrame): Main DataFrame containing all cell coordinates.
    - seg_res (str): Column name representing the segmentation result.
    - num_workers (int): Number of parallel processes to use.
    
    Returns:
    - pd.DataFrame: Concatenated DataFrame with new 'Random_{seg_res}' columns.
    """
    # Get a list of unique cell IDs
    cell_list = coord_df[seg_res].unique()
    
    # Use map instead of submit and wait for better efficiency and simplified code
    with concurrent.futures.ProcessPoolExecutor(max_workers=num_workers) as executor:
        # Use a lambda to pass each cell_id and process it in parallel
        # This approach reduces data transmission by processing subsets efficiently
        results = list(executor.map(
            lambda cell_id: split_cell_with_line(
                coord_df.loc[coord_df[seg_res] == cell_id].copy(),
                seg_res
            ),
            cell_list
        ))
    
    # Concatenate all resulting DataFrames into one
    coord_df_merge = pd.concat(results, axis=0, ignore_index=True)
    return coord_df_merge

def split_cell_with_line(coord_df_sub_cell, seg_res):
	print(coord_df_sub_cell[seg_res].unique())
	cx, cy = coord_df_sub_cell.loc[:,['x', 'y']].mean()
	# generate the angle of the line randomly
	line_angle = np.random.uniform(0, np.pi)
	line_normal = np.array([np.cos(line_angle), np.sin(line_angle)])
	# calculate the the projection from point to line
	points_centered = coord_df_sub_cell.loc[:,['x', 'y']] - np.array([cx, cy])
	projections = np.dot(points_centered, line_normal)
	coord_df_sub_cell['Random_%s' %seg_res] = 1
	coord_df_sub_cell.loc[projections >= 0, 'Random_%s' %seg_res] = 2
	return coord_df_sub_cell

def split_cells_with_line(coord_df, seg_res):
	cell_list = coord_df[seg_res].unique()
	res_list = []
	with concurrent.futures.ProcessPoolExecutor(max_workers = 16) as executor:
		for cell_id in cell_list:
			coord_df_sub_cell = coord_df.loc[coord_df[seg_res] == cell_id, ]
			arg_tuple = (coord_df_sub_cell, seg_res)
			res_list.append(executor.submit(split_cell_with_line, *arg_tuple))
	done, not_done = concurrent.futures.wait(res_list, timeout=None)
	coord_df_list = [future.result() for future in done]
	coord_df_merge = pd.concat(coord_df_list, axis = 0, ignore_index = True)
	return(coord_df_merge)

def cal_corr_within_seg_line(count_df_sub, seg_res, count_name = 'MIDCount', genes_list = None):
	count_df_sub_in_1 = count_df_sub[count_df_sub['Random_%s' %seg_res] == 1]
	count_df_sub_in_2 = count_df_sub[count_df_sub['Random_%s' %seg_res] == 2]
	gene_seg_cell_1 = count_df_sub_in_1[count_name].groupby([count_df_sub_in_1[seg_res], count_df_sub_in_1['geneID']]).sum()
	seg_count_mat_1, seg_count_genes_1, seg_count_cells_1 = longdf_to_mat(gene_seg_cell_1)
	gene_seg_cell_2 = count_df_sub_in_2[count_name].groupby([count_df_sub_in_2[seg_res], count_df_sub_in_2['geneID']]).sum()
	seg_count_mat_2, seg_count_genes_2, seg_count_cells_2 = longdf_to_mat(gene_seg_cell_2)
	if genes_list:
		genes_overlap = list((set(genes_list) & set(seg_count_genes_1)) & set(seg_count_genes_2))
	else:
		genes_overlap = list(set(seg_count_genes_1) & set(seg_count_genes_2))
	cells_overlap = list(set(seg_count_cells_1) & set(seg_count_cells_2))
	seg_count_mat_1_overlap, seg_count_gene_1_overlap, seg_count_cell_1_overlap = sub_mat(mat = seg_count_mat_1, 
								   genes = seg_count_genes_1, 
								   cells = seg_count_cells_1, 
								   cells_sub = cells_overlap, 
								   genes_sub = genes_overlap)
	seg_count_mat_2_overlap, seg_count_gene_2_overlap, seg_count_cell_2_overlap = sub_mat(mat = seg_count_mat_2, 
								   genes = seg_count_genes_2, 
								   cells = seg_count_cells_2, 
								   cells_sub = cells_overlap, 
								   genes_sub = genes_overlap)
	corr_random = pearson_corr(seg_count_mat_1_overlap, seg_count_mat_2_overlap).A.diagonal()
	return(corr_random, seg_count_cell_1_overlap)





# -*- coding: utf-8 -*-
# @Author: dongqing
# @Date:   2024-01-05 20:28:47
# @Last Modified by:   dongqing
# @Last Modified time: 2024-04-17 03:34:46


import os
import warnings
# pd.set_option('mode.chained_assignment', None)
# warnings.filterwarnings('ignore', category=RuntimeWarning)
# warnings.filterwarnings('ignore', category=FutureWarning, append=True)
warnings.filterwarnings('ignore')
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2' 

import spateo as st
import numpy as np

from skimage.io import imread, imsave
from Cellist.Utility import *
from Cellist.IO import *
from Cellist.Plot import *

import matplotlib
import matplotlib.pyplot as plt

plt.rcParams['font.family'] = 'sans-serif'
plt.rcParams['font.sans-serif'] = ['Arial', 'Helvetica']

'''
cellist align --gem /mnt/Venus/home/dongxin/Projects/heart/202401_Stereo_seq/result_C02942B3/041.cellcut/C02942B3.adjusted.gem \
--tif /mnt/Venus/home/dongxin/Projects/heart/202401_Stereo_seq/result_C02942B3/03.register/ssDNA_C02942B3_regist.tif \
--nworkers 24 --outdir .
'''

def AlignmentParser(subparsers):
    parser = subparsers.add_parser("align", 
        help = "Refine alignment between image and spatial transcriptomics. ")
    group_input = parser.add_argument_group("Input arguments")
    group_input.add_argument("--gem", dest = "spot_expr_file", default = None,
        help = "Bin1 gem file generated by SAW pipeline located in '041.cellcut' or '04.tissuecut'. ")
    group_input.add_argument("--tif", dest = "regist_tif_file", default = None,
        help = "Registered tif file generated by SAW pipeline located in '04.tissuecut'. ")
    group_input.add_argument("--nworkers", dest = "num_workers", type = int, default = 8,
        help = "Maximum number of workers to use. ")

    group_output = parser.add_argument_group("Output arguments")
    group_output.add_argument("--outdir", dest = "out_dir", default = '.',
        help = "The output directory. ")
    group_output.add_argument("--outprefix", dest = "out_prefix", default = None,
        help = "The output prefix. ")
    return(parser)

def Alignment(gem_path, img_path, num_workers, out_dir, out_prefix):
    st.config.n_threads = num_workers
    if not out_prefix:
        file_name, file_extension = os.path.splitext(gem_path)
        base_name = os.path.basename(file_name)
        out_gem_prefix = base_name
        file_name, file_extension = os.path.splitext(img_path)
        base_name = os.path.basename(file_name)
        out_img_prefix = base_name
    else:
        out_gem_prefix = out_prefix
        out_img_prefix = out_prefix
    if not os.path.exists(out_dir):
        os.makedirs(out_dir)
    gem_df = pd.read_csv(gem_path, sep = "\t", header = 0)
    x_range = gem_df['x'].max() -  gem_df['x'].min()
    y_range = gem_df['y'].max() -  gem_df['y'].min() 
    # temporary files
    gem_path_spliced = os.path.join(out_dir, "%s_spliced.gem" %out_gem_prefix)
    img = imread(img_path, as_gray = True)
    fig, ax = plt.subplots(1, 1, figsize=(20, 20), sharex=True, sharey=True)
    ax.imshow(img, cmap="gray")
    # ax.set_title("Original", fontsize = 20)
    # plt.tight_layout()
    plt.savefig(os.path.join(out_dir, "%s_original.png" %out_img_prefix), bbox_inches='tight')
    # ratation
    if abs(img.shape[0] - x_range)*abs(img.shape[1] - y_range) > abs(img.shape[0] - y_range)*abs(img.shape[1] - x_range):
        img_path_rotated = os.path.join(out_dir, "%s_transposed.tif" %out_img_prefix)
        rotated = img.T
        rotated = rotated.astype(np.int32)
        imsave(img_path_rotated, rotated)
        fig, ax = plt.subplots(1, 1, figsize=(20, 20), sharex=True, sharey=True)
        ax.imshow(rotated, cmap="gray")
        # ax.set_title("Original", fontsize = 20)
        # plt.tight_layout()
        plt.savefig(os.path.join(out_dir, "%s_transposed.png" %out_img_prefix), bbox_inches='tight')
    else:
        img_path_rotated = img_path
        rotated = img.copy()
        rotated = rotated.astype(np.int32)
    os.system('''sed -e 's/ExonCount/spliced/g' %s > %s''' %(gem_path, gem_path_spliced))
    Spateo_refine_alignment(gem_path_spliced, img_path_rotated, rotated, out_dir, out_img_prefix)

def Spateo_refine_alignment(gem_path, img_path, rotated, out_dir, out_img_prefix):
    # create Spateo class
    adata = st.io.read_bgi_agg(
        gem_path, img_path,
        gene_agg={'nuclear': ['MALAT1', 'NETA1']},
        prealigned = False,
    )
    if 'spliced' in adata.layers:
        adata.layers['unspliced'] = adata.X - adata.layers['spliced']
    else:
        adata.layers['unspliced'] = adata.X
    fig, ax = plt.subplots(ncols=1, figsize=(20, 20), tight_layout=True)
    ax.imshow(adata.layers['stain'])
    st.pl.imshow(adata, 'unspliced', ax=ax, alpha=0.6, cmap='Reds', vmax=3, use_scale=False, save_show_or_return='return')
    ax.set_title('Before alignment', fontsize = 20)
    plt.savefig(os.path.join(out_dir, "%s_before_alignment.png" %out_img_prefix), bbox_inches='tight')
    # refine alignment
    before = adata.layers['stain'].copy()
    st.cs.refine_alignment(adata, mode='rigid', transform_layers=['stain'])
    fig, axes = plt.subplots(ncols=2, figsize=(20, 12), tight_layout=True)
    axes[0].imshow(before)
    st.pl.imshow(adata, 'unspliced', ax=axes[0], alpha=0.6, cmap='Reds', vmax=5, use_scale=False, save_show_or_return='return')
    axes[0].set_title('Before alignment', fontsize = 20)
    st.pl.imshow(adata, 'stain', ax=axes[1], use_scale=False, save_show_or_return='return')
    st.pl.imshow(adata, 'unspliced', ax=axes[1], alpha=0.6, cmap='Reds', vmax=5, use_scale=False, save_show_or_return='return')
    axes[1].set_title('After alignment', fontsize = 20)
    plt.savefig(os.path.join(out_dir, "%s_alignment_compare.png" %out_img_prefix), bbox_inches='tight')
    img_trans = np.pad(adata.layers['stain'], ((rotated.shape[0] - adata.layers['stain'].shape[0],0), (rotated.shape[1] - adata.layers['stain'].shape[1],0)))
    imsave(os.path.join(out_dir, "%s_regist_transposed_aligned_by_Spateo.tiff" %out_img_prefix), img_trans)

if __name__ == '__main__':
    parser = AlignmentParser()
    gem_path = parser.spot_expr_file
    img_path = parser.regist_tif_file
    num_workers = parser.num_workers
    out_dir = parser.out_dir
    out_prefix = parser.out_prefix
    # set number of threads
    Alignment(gem_path, img_path, num_workers, out_dir, out_prefix)




# -*- coding: utf-8 -*-
# @Author: dongqing
# @Date:   2023-09-13 15:02:14
# @Last Modified by:   Dongqing
# @Last Modified time: 2025-08-26 14:02:59


# scale the distance, rather than the coordinates

import gc
import json
import os, sys
import argparse
import functools
import pandas as pd
import numpy as np
import multiprocessing as mp
import seaborn as sns
import skimage
from scipy.spatial import KDTree, distance
from sklearn.preprocessing import StandardScaler
import concurrent.futures
from random import sample, seed
from anndata import ImplicitModificationWarning

from Cellist.Utility import *
from Cellist.IO import *
from Cellist.Plot import *

import logging
logging.getLogger("fontTools.subset").setLevel(logging.WARNING)
logging.basicConfig(level=logging.INFO, format='%(asctime)s  %(message)s')

import warnings
pd.set_option('mode.chained_assignment', None)
warnings.filterwarnings('ignore', category=RuntimeWarning)
warnings.filterwarnings('ignore', category=ImplicitModificationWarning, append=True)

def CellistParser(subparsers):
    parser = subparsers.add_parser("seg", 
        help = "Run Cellist segmentation on high-resolution spatial transcriptomics. ")
    group_input = parser.add_argument_group("Input arguments")
    group_input.add_argument("--platform", dest = "platform", default = 'barcoding', choices = ['barcoding', 'imaging'],
        help = "Supported high-resoultion ST platform: 'barcoding' (e.g., Stereo-seq, Seq-scope) or "
        "'imaging' (e.g., seqFISH, STARmap, 10x Xenium). Default: barcoding.")
    group_input.add_argument("--resolution", dest = "resolution", type = float, default = 0.5,
        help = "Resolution of the platform (in microns), i.e., the distance between spatial units. "
        "For example, in Stereo-seq, where the distance between spots is 0.5 microns, the resolution should be set to 0.5. "
        "In seqFISH+, where the distance between pixels 103 nanometer, it should be set as 0.1. Default: 0.5. ")
    group_input.add_argument("--gem", dest = "spot_expr_file", default = None,
        help = "Bin1 gem file generated by SAW pipeline. ")
    group_input.add_argument("--spot-count-h5", dest = "all_spot_count_h5_file", default = None,
        help = "Bin1 spot-level expression file in h5 format.")
    group_input.add_argument("--nucleus-seg-method", dest = "nucleus_seg_method", default = "Watershed", choices = ['Watershed', 'Cellpose'],
        help = "The initial method to segment nuclei, Watershed or Cellpose. ")
    group_input.add_argument("--nucleus-prop", dest = "props_file", default = None,
        help = "Properties of segmented nucleus by Watershed or Cellpose algorithm. ")
    group_input.add_argument("--nucleus-count-h5", dest = "nucleus_count_h5_file", default = None,
        help = "Nuclei-level expression generated by Watershed or Cellpose segmentation with h5 format. ")
    group_input.add_argument("--nucleus-seg", dest = "nucleus_coord_file", default = None,
        help = "Spot-level Watershed or Cellpose segmentation result. ")
    group_input.add_argument("--nworkers", dest = "num_workers", type = int, default = 8,
        help = "Maximum number of workers to use. ")
    group_input.add_argument("--patch-data-dir", dest = "patch_data_dir", default = None,
        help = "Directory of patch data, for example, 'Data_HVG', generated by Cellist (not used in the first run of Cellist). ")

    group_model = parser.add_argument_group("Model arguments")
    group_model.add_argument("--cell-radius", dest = "max_dist", type = int, default = 15,
        help = "Expected cell radius (in microns). "
        "Considering the diffusion of transcripts, it should be 1.5X~2X larger than the real cell size. "
        "Default: 15")
    group_model.add_argument("--spot-imputation-distance", dest = "neigh_dist", type = float, default = 2.5,
        help = "Imputation distance for each spot (in microns). Default: 2.5. ")
    group_model.add_argument("--noise-prop", dest = "noise_prop", type = float, default = 0.25,
        help = "Proportion of noise spots. Default: 0.25. ")
    group_model.add_argument("--alpha", dest = "alpha", type = float, default = 0.8,
        help = "Weight for expression distance. Default: 0.8. ")
    group_model.add_argument("--sigma", dest = "sigma", type = float, default = 1.0,
        help = "Weight for regularization. Default: 1.0. ")
    group_model.add_argument("--beta", dest = "beta", type = float, default = 10,
        help = "Penalty weight. Default: 10. ")
    group_model.add_argument("--gene-use", dest = "gene_use", type = str, default = 'HVG', choices = ['HVG', 'Frequent'],
        help = "Gene to use to calculate spot-cell distance. "
        "'HVG' means highly variable genes, and 'Frequent' means top frequent genes. "
        "Default: HVG. ")
    group_model.add_argument("--two-step", dest = "two_step", action = "store_true",
        help = "Two-step mode. If set, Cellist will assign non-nuclei spots recusively. ")
    group_model.add_argument("--cyto", dest = "cyto", action = "store_true", 
        help = "Two-step cyto mode. If set, Cellist will calculate the correlation between non-nuclei spots and the 1st cycle of cytoplasm. ")
    group_model.add_argument("--cell-radius-s1-scale", dest = "max_dist_s1_scale",type = float,  default = 0.5,
        help = "Two-step mode. The allowed maximum distance scale in the first-step segmentation. ")
    group_model.add_argument("--noise-prop-s1", dest = "noise_prop_s1", type = float,  default = 0.4,
        help = "Proportion of noise spots in the first step. Default: 0.4. ")

    group_output = parser.add_argument_group("Output arguments")
    group_output.add_argument("--outdir", dest = "out_dir", default = "",help = "Output directory.")
    group_output.add_argument("--outprefix", dest = "out_prefix", default = "",help = "Output prefix.")
    return(parser)

def Update_centroid(phy_coord_df, cluster_col):
    phy_centroid = phy_coord_df.groupby([cluster_col]).mean()
    return(phy_centroid)

def Update_cell_expr(coord_df_slice, count_df_slice, count_name):
    count_df_slice = pd.merge(count_df_slice, coord_df_slice.loc[:, ('x_y', 'Cellist', 'Nucleus')], how = "inner", on = 'x_y')
    expr_mat, gene_list, cell_list = get_cell_mat(count_df_seg = count_df_slice, seg_res = 'Cellist', count_name = count_name)
    return(expr_mat, gene_list, cell_list)

def Assign_nonnucleispot(cell_expr_slice_array, phy_cell_centroid, nonnucl_enhanced_array_filtered, nonnucl_phy_coord_df,
    coord_df_slice, max_dist_slice, alpha, sigma, beta, noise_prop):
    nonnucl_spot_centroid_corr = pearson_corr(nonnucl_enhanced_array_filtered, cell_expr_slice_array)
    nonnucl_spot_centroid_corr[np.isnan(nonnucl_spot_centroid_corr)] = -1
    nonnucl_spot_centroid_dist_phy = distance.cdist(nonnucl_phy_coord_df, phy_cell_centroid, 'euclidean')
    # dist_scale_factor = np.max(nonnucl_spot_centroid_dist_phy)
    dist_scale_factor = 2*max_dist_slice
    P_dist = (nonnucl_spot_centroid_dist_phy - max_dist_slice)/dist_scale_factor
    P_dist[P_dist < 0] = 0
    R = np.exp(-(alpha*(1-nonnucl_spot_centroid_corr)/2 + (1-alpha)*nonnucl_spot_centroid_dist_phy/dist_scale_factor + beta*np.square(P_dist))/sigma)
    expr_loss = alpha*(1-nonnucl_spot_centroid_corr)/2
    dist_loss = (1-alpha)*nonnucl_spot_centroid_dist_phy/np.max(nonnucl_spot_centroid_dist_phy)
    regularization = sigma*np.log(R + 0.01)
    penalty = beta*np.square(P_dist)
    loss = expr_loss + dist_loss + regularization + penalty
    obj = np.nansum(np.multiply(R, loss))
    nonnucl_phy_coord_df.loc[:, 'Cellist'] = phy_cell_centroid.index[np.argmax(R, axis = 1)]
    nonnucl_phy_coord_df.loc[:, 'Cellist_prob'] = R.max(axis = 1)
    # nonnucl_phy_coord_df.loc[:, 'Cellist_slice'] = nonnucl_phy_coord_df.loc[:, 'Cellist'].copy()
    # q_noise = nonnucl_phy_coord_df['Cellist_prob'].quantile(noise_prop)
    # nonnucl_phy_coord_df.loc[nonnucl_phy_coord_df['Cellist_prob'] < q_noise, 'Cellist_slice'] = np.nan
    # nonnucl_spot_centroid_dist_phy_assigned = np.array([nonnucl_spot_centroid_dist_phy[i, np.argmax(R, axis = 1)[i]] for i in range(nonnucl_spot_centroid_dist_phy.shape[0])])
    # nonnucl_phy_coord_df.loc[nonnucl_spot_centroid_dist_phy_assigned > 2*max_dist_slice, 'Cellist_slice'] = np.nan
    R_df = pd.DataFrame(R, index = nonnucl_phy_coord_df.index, columns = phy_cell_centroid.index)
    R_sum = np.array(np.nansum(R, axis = 1).tolist()*R.shape[1]).reshape(R.shape, order = 'F')
    R_norm = np.divide(R, R_sum)
    R_norm[np.isnan(R_norm)] = 0
    nonnucl_phy_coord_df.loc[:, 'Cellist_prob_norm'] = R_norm.max(axis = 1)
    R_norm_df = pd.DataFrame(R_norm, index = nonnucl_phy_coord_df.index, columns = phy_cell_centroid.index)
    return(R_df, obj, nonnucl_phy_coord_df)

def Soft_assignment(nucleus_seg_method, nonnucl_enhanced_array, nonnucl_enhanced_genes, nonnucl_enhanced_spots, 
    coord_df_slice, count_df_slice, 
    max_dist_slice, alpha, sigma, beta, noise_prop):
    coord_df_slice.loc[:, 'Cellist'] = coord_df_slice[nucleus_seg_method].tolist()
    # not update cell centroid
    phy_cell_centroid = Update_centroid(phy_coord_df = coord_df_slice.loc[:, ('x', 'y', nucleus_seg_method)], cluster_col = nucleus_seg_method)
    count_name = count_df_slice.columns[3]
    try:
        cell_expr_slice_array, cell_expr_slice_gene, cell_expr_slice_cell = Update_cell_expr(coord_df_slice = coord_df_slice, count_df_slice = count_df_slice, count_name = count_name)
    except:
        return(None)
    gene_overlap = sorted(list(set(cell_expr_slice_gene) & set(nonnucl_enhanced_genes)))
    cell_expr_slice_array_filtered, cell_expr_slice_gene_filtered, cell_expr_slice_cell_filtered = sub_mat(
        mat = cell_expr_slice_array, genes = cell_expr_slice_gene, cells = cell_expr_slice_cell, 
        genes_sub = gene_overlap)
    nonnucl_enhanced_array_filtered, nonnucl_enhanced_genes_filtered, nonnucl_enhanced_spots_filtered = sub_mat(
        mat = nonnucl_enhanced_array, genes = nonnucl_enhanced_genes, cells = nonnucl_enhanced_spots, 
        genes_sub = cell_expr_slice_gene_filtered)
    coord_df_slice_nonnucl_df = coord_df_slice.loc[nonnucl_enhanced_spots_filtered, ('x', 'y')]
    phy_cell_centroid = phy_cell_centroid.loc[[float(i) for i in cell_expr_slice_cell_filtered],:]
    R_df, obj, nonnucl_phy_coord_df = Assign_nonnucleispot(
        cell_expr_slice_array = cell_expr_slice_array_filtered.toarray(),
        phy_cell_centroid = phy_cell_centroid, 
        nonnucl_enhanced_array_filtered = nonnucl_enhanced_array_filtered,
        nonnucl_phy_coord_df = coord_df_slice_nonnucl_df,
        coord_df_slice = coord_df_slice, max_dist_slice = max_dist_slice,
        alpha = alpha, sigma = sigma, beta = beta, noise_prop = noise_prop)
    coord_df_slice.loc[nonnucl_phy_coord_df.index.tolist(), "Cellist"] = nonnucl_phy_coord_df['Cellist'].tolist()
    coord_df_slice.loc[:, 'Cellist_prob'] = 1
    coord_df_slice.loc[nonnucl_phy_coord_df.index.tolist(), "Cellist_prob"] = nonnucl_phy_coord_df['Cellist_prob'].tolist()
    coord_df_slice.loc[coord_df_slice['Cellist_prob'] == 0.0, "Cellist"] = np.nan
    # coord_df_slice.loc[:, 'Cellist_slice'] = coord_df_slice.loc[:, 'Cellist'].copy()
    # coord_df_slice.loc[nonnucl_phy_coord_df.index.tolist(), "Cellist_slice"] = nonnucl_phy_coord_df['Cellist_slice'].tolist()
    coord_df_slice.loc[:, 'Cellist_prob_norm'] = 1
    coord_df_slice.loc[nonnucl_phy_coord_df.index.tolist(), "Cellist_prob_norm"] = nonnucl_phy_coord_df['Cellist_prob_norm'].tolist()
    coord_df_slice.loc[coord_df_slice['Cellist_prob_norm'] == 0.0, "Cellist_norm"] = np.nan
    return(coord_df_slice)

def Soft_assignment_2step(nucleus_seg_method, nonnucl_enhanced_array, nonnucl_enhanced_genes, nonnucl_enhanced_spots, 
    coord_df_slice, count_df_slice,
    max_dist_slice, alpha, sigma, beta, noise_prop, max_dist_s1_scale = 0.5, noise_prop_s1 = 0.4):
    coord_df_slice.loc[:, 'Cellist'] = coord_df_slice[nucleus_seg_method].tolist()
    # not update cell centroid
    multi_step = 2
    for i in range(multi_step):
        print('Segmentation %s.' %i)
        phy_cell_centroid = Update_centroid(phy_coord_df = coord_df_slice.loc[:, ('x', 'y', 'Cellist')], cluster_col = 'Cellist')
        count_name = count_df_slice.columns[3]
        try:
            cell_expr_slice_array, cell_expr_slice_gene, cell_expr_slice_cell = Update_cell_expr(coord_df_slice = coord_df_slice, count_df_slice = count_df_slice, count_name = count_name)
        except:
            return(None)
        gene_overlap = sorted(list(set(cell_expr_slice_gene) & set(nonnucl_enhanced_genes)))
        cell_expr_slice_array_filtered, cell_expr_slice_gene_filtered, cell_expr_slice_cell_filtered = sub_mat(
            mat = cell_expr_slice_array, genes = cell_expr_slice_gene, cells = cell_expr_slice_cell, 
            genes_sub = gene_overlap)
        coord_df_slice_nonnucl = coord_df_slice.loc[coord_df_slice['Cellist'].isna(),:]
        nonnucl_spots = sorted(list(set(nonnucl_enhanced_spots) & set(coord_df_slice_nonnucl.index.tolist())))
        nonnucl_enhanced_array_filtered, nonnucl_enhanced_genes_filtered, nonnucl_enhanced_spots_filtered = sub_mat(
            mat = nonnucl_enhanced_array, genes = nonnucl_enhanced_genes, cells = nonnucl_enhanced_spots, 
            genes_sub = cell_expr_slice_gene_filtered, cells_sub = nonnucl_spots)
        coord_df_slice_nonnucl_df = coord_df_slice.loc[nonnucl_enhanced_spots_filtered, ('x', 'y')]
        phy_cell_centroid = phy_cell_centroid.loc[[float(i) for i in cell_expr_slice_cell_filtered],:]
        if i == 0:
            max_dist_slice_step = max_dist_slice*max_dist_s1_scale
        else:
            max_dist_slice_step = max_dist_slice
        print(coord_df_slice_nonnucl_df.shape[0])
        R_df, obj, nonnucl_phy_coord_df = Assign_nonnucleispot(
            cell_expr_slice_array = cell_expr_slice_array_filtered.toarray(),
            phy_cell_centroid = phy_cell_centroid, 
            nonnucl_enhanced_array_filtered = nonnucl_enhanced_array_filtered,
            nonnucl_phy_coord_df = coord_df_slice_nonnucl_df,
            coord_df_slice = coord_df_slice, max_dist_slice = max_dist_slice_step,
            alpha = alpha, sigma = sigma, beta = beta, noise_prop = noise_prop)
        coord_df_slice.loc[nonnucl_phy_coord_df.index.tolist(), "Cellist"] = nonnucl_phy_coord_df['Cellist'].tolist()
        coord_df_slice.loc[:, 'Cellist_prob'] = 1
        coord_df_slice.loc[nonnucl_phy_coord_df.index.tolist(), "Cellist_prob"] = nonnucl_phy_coord_df['Cellist_prob'].tolist()
        coord_df_slice.loc[coord_df_slice['Cellist_prob'] == 0.0, "Cellist"] = np.nan
        coord_df_slice.loc[:, 'Cellist_prob_norm'] = 1
        coord_df_slice.loc[nonnucl_phy_coord_df.index.tolist(), "Cellist_prob_norm"] = nonnucl_phy_coord_df['Cellist_prob_norm'].tolist()
        coord_df_slice.loc[coord_df_slice['Cellist_prob_norm'] == 0.0, "Cellist_norm"] = np.nan
        if i == 0:
            coord_df_slice_cyto = coord_df_slice.loc[coord_df_slice['Nucleus'] == 0, ]
            q_noise = coord_df_slice_cyto['Cellist_prob'].quantile(noise_prop_s1)
            coord_df_slice.loc[coord_df_slice['Cellist_prob'] < q_noise, 'Cellist'] = np.nan
    return(coord_df_slice)

def Soft_assignment_2step_cyto(nucleus_seg_method, nonnucl_enhanced_array, nonnucl_enhanced_genes, nonnucl_enhanced_spots, 
    coord_df_slice_scaled_df, coord_df_slice, count_df_slice, 
    max_dist_slice_scaled, alpha, sigma, beta, prob_cutoff):
    coord_df_slice.loc[:, 'Cellist'] = coord_df_slice[nucleus_seg_method].tolist()
    multi_step = 2
    for i in range(multi_step):
        # cyto1-nucl
        print('Segmentation %s.' %i)
        coord_df_slice_scaled_df.loc[:,'Cellist'] = coord_df_slice.loc[:,'Cellist']
        phy_cell_centroid = Update_centroid(phy_coord_df = coord_df_slice_scaled_df.loc[:, ('x', 'y', 'Cellist')], cluster_col = 'Cellist')
        count_name = count_df_slice.columns[3]
        try:
            if i > 0:
                coord_df_slice_cyto1 = coord_df_slice.loc[(coord_df_slice['Cellist'].notna()) & (coord_df_slice['Nucleus'] == 0),:]
                cell_expr_slice_array, cell_expr_slice_gene, cell_expr_slice_cell = Update_cell_expr(coord_df_slice = coord_df_slice_cyto1, count_df_slice = count_df_slice, count_name = count_name)
            else:
                cell_expr_slice_array, cell_expr_slice_gene, cell_expr_slice_cell = Update_cell_expr(coord_df_slice = coord_df_slice, count_df_slice = count_df_slice, count_name = count_name)
        except:
            return(None)
        gene_overlap = sorted(list(set(cell_expr_slice_gene) & set(nonnucl_enhanced_genes)))
        cell_expr_slice_array_filtered, cell_expr_slice_gene_filtered, cell_expr_slice_cell_filtered = sub_mat(
            mat = cell_expr_slice_array, genes = cell_expr_slice_gene, cells = cell_expr_slice_cell, 
            genes_sub = gene_overlap)
        coord_df_slice_nonnucl = coord_df_slice.loc[coord_df_slice['Cellist'].isna(),:]
        nonnucl_spots = sorted(list(set(nonnucl_enhanced_spots) & set(coord_df_slice_nonnucl.index.tolist())))
        nonnucl_enhanced_array_filtered, nonnucl_enhanced_genes_filtered, nonnucl_enhanced_spots_filtered = sub_mat(
            mat = nonnucl_enhanced_array, genes = nonnucl_enhanced_genes, cells = nonnucl_enhanced_spots, 
            genes_sub = cell_expr_slice_gene_filtered, cells_sub = nonnucl_spots)
        coord_df_slice_nonnucl_scaled_df = coord_df_slice_scaled_df.loc[nonnucl_enhanced_spots_filtered, ('x', 'y')]
        phy_cell_centroid = phy_cell_centroid.loc[[float(i) for i in cell_expr_slice_cell_filtered],:]
        R_norm, obj, nonnucl_phy_coord_df = Assign_nonnucleispot(
            cell_expr_slice_array = cell_expr_slice_array_filtered.toarray(),
            phy_cell_centroid = phy_cell_centroid, 
            nonnucl_enhanced_array_filtered = nonnucl_enhanced_array_filtered,
            nonnucl_phy_coord_df = coord_df_slice_nonnucl_scaled_df,
            coord_df_slice = coord_df_slice, max_dist_slice_scaled = max_dist_slice_scaled/(multi_step-i),
            alpha = alpha, sigma = sigma, beta = beta, prob_cutoff = prob_cutoff)
        coord_df_slice.loc[nonnucl_phy_coord_df.index.tolist(), "Cellist"] = nonnucl_phy_coord_df['Cellist'].tolist()
        coord_df_slice.loc[:, 'Cellist_prob'] = 1
        coord_df_slice.loc[nonnucl_phy_coord_df.index.tolist(), "Cellist_prob"] = nonnucl_phy_coord_df['Cellist_prob'].tolist()
        # coord_df_slice.loc[coord_df_slice['Cellist_prob'].isna(), "Cellist"] = np.nan
        coord_df_slice.loc[coord_df_slice['Cellist_prob'] == 0.0, "Cellist"] = np.nan
    return(coord_df_slice)

def Soft_assignment_slice(count_df_patch, coord_df_patch, nucleus_seg_method, props_df_patch,
    all_count_mat, all_count_genes, all_count_spots, 
    xmin_cur, xmax_cur, ymin_cur, ymax_cur, selected_genes, max_dist = 15, resolution = 0.5,
    alpha = 0.8, out_dir = ".", sigma = 1.0, beta = 10, two_step = False, cyto = False, max_dist_s1_scale = 0.5, noise_prop_s1 = 0.4, noise_prop = 0.25, neigh_dist = 2.5):
    coord_df_slice = coord_df_patch.loc[(coord_df_patch['x'] >= xmin_cur) & (coord_df_patch['x'] < xmax_cur) &
                                (coord_df_patch['y'] >= ymin_cur) & (coord_df_patch['y'] < ymax_cur) , :]
    count_df_slice = count_df_patch.loc[(count_df_patch['x'] >= xmin_cur) & (count_df_patch['x'] < xmax_cur) &
                                (count_df_patch['y'] >= ymin_cur) & (count_df_patch['y'] < ymax_cur) , :]
    coord_df_slice_nonnucl = coord_df_slice[coord_df_slice['Nucleus'] == 0]
    coord_df_slice_nucl = coord_df_slice[coord_df_slice['Nucleus'] == 1]
    nucl_spot_count = coord_df_slice_nucl[nucleus_seg_method].value_counts()
    props_df_slice = props_df_patch.loc[(props_df_patch['centroid-0'] >= xmin_cur) & (props_df_patch['centroid-0'] < xmax_cur) & 
                                (props_df_patch['centroid-1'] >= ymin_cur) & (props_df_patch['centroid-1'] < ymax_cur), :]
    if coord_df_slice_nucl.shape[0] > 0 and props_df_slice.shape[0] > 0:
        props_df_slice.loc[np.intersect1d(props_df_slice.index, nucl_spot_count.index),"nSpot"] = nucl_spot_count[np.intersect1d(props_df_slice.index, nucl_spot_count.index)]
        props_df_slice = props_df_slice.loc[props_df_slice['nSpot'] >= 20, :]
        if props_df_slice.shape[0] > 0:
            nucl_cell_slice = list(set(coord_df_slice_nucl[nucleus_seg_method].unique()) & set(props_df_slice.index.tolist()))
            nucl_cell_slice = sorted(nucl_cell_slice)
            nucl_cell_rm = set(coord_df_slice_nucl[nucleus_seg_method].unique()) - set(nucl_cell_slice)
            if nucl_cell_slice:
                coord_df_slice.loc[coord_df_slice[nucleus_seg_method].isin(nucl_cell_rm), nucleus_seg_method] = np.nan
                spot_expr_slice_filter, spot_gene_slice_filter, spot_slice_filter = sub_mat(
                    mat = all_count_mat, genes = all_count_genes, cells = all_count_spots, 
                    cells_sub = coord_df_slice.index.tolist(), 
                    genes_sub = selected_genes)
                nonnucl_spots = sorted(list(set(spot_slice_filter) & set(coord_df_slice_nonnucl.index.tolist())))
                if nonnucl_spots:
                    # enhance single spot's expr
                    neigh_dist_slice = int(neigh_dist/resolution)
                    spot_expr_slice_filter_enhanced_array = enhance_spot_expr(coord_df = coord_df_slice.loc[spot_slice_filter, :], 
                        coord_expr_mat = spot_expr_slice_filter, neigh_dist = neigh_dist_slice)
                    spot_expr_slice_filter_enhanced_array_nonnucl, spot_expr_slice_filter_enhanced_array_nonnucl_gene, spot_expr_slice_filter_enhanced_array_nonnucl_spot = sub_mat(
                        mat = spot_expr_slice_filter_enhanced_array,
                        genes = spot_gene_slice_filter, 
                        cells = spot_slice_filter,
                        cells_sub = nonnucl_spots)
                    coord_df_slice_nonnucl = coord_df_slice_nonnucl.loc[spot_expr_slice_filter_enhanced_array_nonnucl_spot, :]
                    if type(max_dist) == int:
                        max_dist_slice = max_dist/resolution
                    else: # nuclei size
                        max_dist_slice = 1.5*props_df_slice['equivalent_diameter_area'].median()
                    if two_step:
                        if not cyto:
                            coord_df_slice = Soft_assignment_2step(
                                nucleus_seg_method = nucleus_seg_method,
                                nonnucl_enhanced_array = spot_expr_slice_filter_enhanced_array_nonnucl,
                                nonnucl_enhanced_genes = spot_expr_slice_filter_enhanced_array_nonnucl_gene,
                                nonnucl_enhanced_spots = spot_expr_slice_filter_enhanced_array_nonnucl_spot,
                                coord_df_slice = coord_df_slice,
                                count_df_slice = count_df_slice, 
                                max_dist_slice = max_dist_slice,
                                alpha = alpha, sigma = sigma, beta = beta, 
                                noise_prop = noise_prop,
                                max_dist_s1_scale = max_dist_s1_scale, 
                                noise_prop_s1 = noise_prop_s1)
                        else:
                            coord_df_slice = Soft_assignment_2step_cyto(
                                nucleus_seg_method = nucleus_seg_method,
                                nonnucl_enhanced_array = spot_expr_slice_filter_enhanced_array_nonnucl,
                                nonnucl_enhanced_genes = spot_expr_slice_filter_enhanced_array_nonnucl_gene,
                                nonnucl_enhanced_spots = spot_expr_slice_filter_enhanced_array_nonnucl_spot,
                                coord_df_slice_scaled_df = coord_df_slice_scaled_df, 
                                coord_df_slice = coord_df_slice,
                                count_df_slice = count_df_slice, 
                                max_dist_slice_scaled = max_dist_slice_scaled,
                                alpha = alpha, sigma = sigma, beta = beta, 
                                prob_cutoff = prob_cutoff)
                    else:
                        coord_df_slice = Soft_assignment(
                            nucleus_seg_method = nucleus_seg_method,
                            nonnucl_enhanced_array = spot_expr_slice_filter_enhanced_array_nonnucl,
                            nonnucl_enhanced_genes = spot_expr_slice_filter_enhanced_array_nonnucl_gene,
                            nonnucl_enhanced_spots = spot_expr_slice_filter_enhanced_array_nonnucl_spot,
                            coord_df_slice = coord_df_slice,
                            count_df_slice = count_df_slice, 
                            max_dist_slice = max_dist_slice,
                            alpha = alpha, sigma = sigma, beta = beta, noise_prop = noise_prop)
                    if type(coord_df_slice) == pd.DataFrame:
                        tmp_dir = os.path.join(out_dir, "tmp")
                        out_file_tmp = os.path.join(tmp_dir, "Sub_%s_%s_%s_%s_soft_clustering.txt" %(xmin_cur, xmax_cur, ymin_cur, ymax_cur))
                        coord_df_slice.to_csv(out_file_tmp, index = False, sep = "\t")
                    else:
                        print("%s:%s:%s:%s:empty" %(xmin_cur, xmax_cur, ymin_cur, ymax_cur))
                        out_file_tmp = "NULL"
                else:
                    coord_df_slice['Cellist'] = coord_df_slice[nucleus_seg_method]
                    coord_df_slice['Cellist_prob'] = 1
                    tmp_dir = os.path.join(out_dir, "tmp")
                    out_file_tmp = os.path.join(tmp_dir, "Sub_%s_%s_%s_%s_soft_clustering.txt" %(xmin_cur, xmax_cur, ymin_cur, ymax_cur))
                    coord_df_slice.to_csv(out_file_tmp, index = False, sep = "\t")
            else:
                print("%s:%s:%s:%s:empty" %(xmin_cur, xmax_cur, ymin_cur, ymax_cur))
                out_file_tmp = "NULL"
        else:
            print("%s:%s:%s:%s:empty" %(xmin_cur, xmax_cur, ymin_cur, ymax_cur))
            out_file_tmp = "NULL"
    else:
        print("%s:%s:%s:%s:empty" %(xmin_cur, xmax_cur, ymin_cur, ymax_cur))
        out_file_tmp = "NULL"
    return(out_file_tmp)


def Write_patch(count_df, coord_df, nucleus_seg_method, props_df, nucleus_expr_mat, nucleus_gene, nucleus_cell, all_count_mat, all_count_genes, all_count_spots,
    xmin_cur, xmax_cur, ymin_cur, ymax_cur,
    patch_data_dir, gene_use = 'HVG'):
    # get small patches to set parameters
    coord_df_patch = coord_df.loc[(coord_df['x'] >= xmin_cur) & (coord_df['x'] < xmax_cur) &
                                (coord_df['y'] >= ymin_cur) & (coord_df['y'] < ymax_cur) , :]
    coord_df_patch_nucl = coord_df_patch[coord_df_patch['Nucleus'] == 1]
    props_df_patch = props_df.loc[(props_df['centroid-0'] >= xmin_cur) & (props_df['centroid-0'] < xmax_cur) & 
                                (props_df['centroid-1'] >= ymin_cur) & (props_df['centroid-1'] < ymax_cur), :]
    count_df_patch = count_df.loc[(count_df['x'] >= xmin_cur) & (count_df['x'] < xmax_cur) &
                                (count_df['y'] >= ymin_cur) & (count_df['y'] < ymax_cur) , :]
    patch_prefix = "Patch_%s_%s_%s_%s" %(xmin_cur, xmax_cur, ymin_cur, ymax_cur)
    if coord_df_patch_nucl.shape[0] > 0 and props_df_patch.shape[0] > 0:
        nucl_cell_patch = list(set(coord_df_patch_nucl[nucleus_seg_method].unique()) & set(props_df_patch.index.tolist()))
        nucl_cell_patch = sorted(nucl_cell_patch)
        logging.info('There are %s cells in this patch.' %(len(nucl_cell_patch)))
        if len(nucl_cell_patch) > 0:
            nucleus_expr_sub, nucleus_gene_sub, nucleus_cell_sub = sub_mat(mat = nucleus_expr_mat, genes = nucleus_gene, cells = nucleus_cell, 
                cells_sub = nucl_cell_patch)
            if gene_use == "HVG":
                # get highly varibale genes
                try:
                    selected_genes = get_hvg(nucleus_expr_sub, nucleus_gene_sub, nucleus_cell_sub, 1500)
                except:
                    logging.info('Fail to identify HVG in %s.' %patch_prefix)
                    return('NULL')
            else:
                selected_genes = get_frequent_gene(nucleus_expr_sub, nucleus_gene_sub, nucleus_cell_sub, 0.05)
            del nucleus_expr_sub, nucleus_gene_sub, nucleus_cell_sub
            gc.collect()
            try:
                spot_expr_patch_filter, spot_gene_patch_filter, spot_patch = sub_mat(
                        mat = all_count_mat, genes = all_count_genes, cells = all_count_spots, 
                        cells_sub = coord_df_patch.index.tolist(), 
                        genes_sub = selected_genes)
            except:
                sys.exit('Fail to subset matrix in %s.' %patch_prefix)
            Write_tmp_files(count_df_patch, coord_df_patch, props_df_patch, spot_expr_patch_filter, spot_gene_patch_filter, spot_patch, patch_prefix, patch_data_dir)
            return(patch_prefix)
        else:
            return('NULL')
    else:
        return('NULL')

def Soft_assignment_patch(platform, resolution, nucleus_seg_method, props_file_patch, coord_file_patch, all_spot_count_h5_file_patch, count_file_patch,
    out_dir, num_workers, 
    alpha, sigma, beta, gene_use, max_dist, two_step, cyto, max_dist_s1_scale, noise_prop_s1, noise_prop, neigh_dist):
    props_df_patch = pd.read_csv(props_file_patch, sep = "\t", header = 0, index_col = 0)
    coord_df_patch = pd.read_csv(coord_file_patch, sep = "\t", header = 0, index_col = 0)
    count_df_patch = pd.read_csv(count_file_patch, sep = "\t", header = 0)
    spot_expr_patch_filter, spot_gene_patch_filter, spot_patch = read_h5(all_spot_count_h5_file_patch)
    patch_prefix = os.path.basename(props_file_patch).split("Patch_")[1].split('_property.txt')[0]
    if platform == "barcoding":
        slice_size = 200
        stride_length = 10
    elif platform == "imaging":
        slice_size = 2200
        stride_length = 100
    selected_genes = spot_gene_patch_filter
    coord_slice_list = slice_coord(coord_df_patch, slice_size, stride_length)
    logging.info('There are %s slices in the patch.' %len(coord_slice_list))
    res_list = []
    with concurrent.futures.ThreadPoolExecutor(max_workers = 36) as executor:
        for coord_tuple in coord_slice_list:
            xmin_slice, xmax_slice, ymin_slice, ymax_slice = coord_tuple
            arg_tuple = (count_df_patch, coord_df_patch, nucleus_seg_method, props_df_patch, spot_expr_patch_filter, 
                spot_gene_patch_filter, spot_patch, xmin_slice, xmax_slice, ymin_slice, ymax_slice, 
                selected_genes, max_dist, resolution, alpha, out_dir, sigma, beta, two_step, cyto,max_dist_s1_scale, noise_prop_s1, noise_prop, neigh_dist)
            res_list.append(executor.submit(Soft_assignment_slice, *arg_tuple))
        done, not_done = concurrent.futures.wait(res_list, timeout=None)
        tmp_file_list = [future.result() for future in done]
    return(tmp_file_list)


def Read_files(nucleus_seg_method, props_file, nucleus_count_h5_file, nucleus_coord_file, all_spot_count_h5_file, spot_expr_file):
    # ---------- Read files ------------
    # Read property file
    props_df = pd.read_csv(props_file, sep = "\t")
    props_df.index = props_df['label'].astype(np.float64)
    props_df.index.name = None
    # Read nucleus expr file
    nucleus_expr_mat, nucleus_gene, nucleus_cell = read_h5(nucleus_count_h5_file)
    nucleus_cell = [float(i) for i in nucleus_cell]
    # Read nucleus file
    coord_df = pd.read_csv(nucleus_coord_file, sep = "\t", header = 0)
    coord_df.index = coord_df['x_y']
    coord_df.index.name = None
    coord_df.loc[coord_df['Nucleus'] == 0, nucleus_seg_method] = np.nan
    # Read spot-level expr file (with nucleus 0/1)]
    count_df = pd.read_csv(spot_expr_file, sep = "\t", comment = "#")
    count_df.loc[:, 'x_y'] = count_df['x'].astype(str) + '_' + count_df['y'].astype(str)
    # Read spot-level expr mat
    all_count_mat, all_count_genes, all_count_spots = read_h5(all_spot_count_h5_file)
    return(count_df, coord_df, props_df, 
        nucleus_expr_mat, nucleus_gene, nucleus_cell, 
        all_count_mat, all_count_genes, all_count_spots)

def Write_tmp_files(count_df_patch, coord_df_patch, props_df_patch, spot_expr_patch_filter, spot_gene_patch_filter, spot_patch, 
    patch_prefix, patch_data_dir):
    # ---------- Write tmp patch files ------------
    # Write property file
    props_df_patch.to_csv(os.path.join(patch_data_dir, "%s_property.txt" %patch_prefix), sep = "\t")
    # Write spot coord file
    coord_df_patch.to_csv(os.path.join(patch_data_dir, "%s_coord.txt" %patch_prefix), sep = "\t")
    # Write spot-level expr file (with nucleus 0/1)]
    count_df_patch.to_csv(os.path.join(patch_data_dir, "%s_count.txt" %patch_prefix), sep = "\t", index = False)
    # Write spot-level expr mat
    write_10X_h5(os.path.join(patch_data_dir, "%s_spot_count.h5" %patch_prefix), spot_expr_patch_filter, spot_gene_patch_filter, spot_patch, datatype = 'Gene')

def Assign_filter(coord_df, out_dir, out_prefix, noise_prop = 0.25):
    coord_df_nucl = coord_df.loc[coord_df['Nucleus']==1, ]
    coord_df_cyto = coord_df.loc[coord_df['Nucleus']==0, ]
    # hist_res = plt.hist(coord_df_cyto['Cellist_prob'], bins = 200)
    # hist_center = (hist_res[1][:-1] + hist_res[1][1:])/2
    # otsu2_thre = skimage.filters.threshold_otsu(hist = (hist_res[0], hist_center))
    # otsu3_thre = skimage.filters.threshold_multiotsu(classes=3, hist = (hist_res[0], hist_center))
    f, ax = plt.subplots(1,1, figsize=(4, 1.6))
    line1 = sns.kdeplot(data = coord_df_cyto, 
        x = "Cellist_prob", ax = ax, color = "#C00000")
    line2= ax.axvline(x = coord_df_cyto['Cellist_prob'].quantile(0.25), color="black", linestyle="--", linewidth = 0.9)
    # line3= ax.axvline(x = otsu2_thre, color="#009292", linestyle=":", linewidth = 1)
    # line4= ax.axvline(x = otsu3_thre[0], color="#FFB6DB", linestyle="-", linewidth = 1)
    # line5= ax.axvline(x = otsu3_thre[1], color="#FF6DB6", linestyle="-", linewidth = 1)
    # plt.legend(handles = [line2, line3, line4, line5], labels = ['Q1', 'Otsu threshold (binary)', '3-class, lower', '3-class, upper'],
    #     fontsize = 7.5, markerscale = 0.8)
    plt.xlabel('Assign probability')
    plt.title(out_prefix, pad = 10)
    plot_file = os.path.join(out_dir, "%s_assign_probability.pdf" %out_prefix)
    plt.savefig(plot_file, bbox_inches='tight')
    # norm
    # hist_res = plt.hist(coord_df_cyto['Cellist_prob_norm'], bins = 200)
    # hist_center = (hist_res[1][:-1] + hist_res[1][1:])/2
    # otsu2_thre = skimage.filters.threshold_otsu(hist = (hist_res[0], hist_center))
    # otsu3_thre = skimage.filters.threshold_multiotsu(classes=3, hist = (hist_res[0], hist_center))
    f, ax = plt.subplots(1,1, figsize=(4, 1.6))
    line1 = sns.kdeplot(data = coord_df_cyto, 
        x = "Cellist_prob_norm", ax = ax, color = "#C00000")
    line2= ax.axvline(x = coord_df_cyto['Cellist_prob_norm'].quantile(0.25), color="black", linestyle="--", linewidth = 0.9)
    # line3= ax.axvline(x = otsu2_thre, color="#009292", linestyle=":", linewidth = 1)
    # line4= ax.axvline(x = otsu3_thre[0], color="#FFB6DB", linestyle="-", linewidth = 1)
    # line5= ax.axvline(x = otsu3_thre[1], color="#FF6DB6", linestyle="-", linewidth = 1)
    # plt.legend(handles = [line2, line3, line4, line5], labels = ['Q1', 'Otsu threshold (binary)', '3-class, lower', '3-class, upper'],
    #     fontsize = 7.5, markerscale = 0.8)
    plt.xlabel('Assign probability')
    plt.title(out_prefix, pad = 10)
    plot_file = os.path.join(out_dir, "%s_assign_probability_normalized.pdf" %out_prefix)
    plt.savefig(plot_file, bbox_inches='tight')
    q_noise = coord_df_cyto['Cellist_prob'].quantile(noise_prop)
    coord_df['Cellist_unfiltered'] = coord_df['Cellist'].copy()
    coord_df.loc[coord_df['Cellist_prob'] < q_noise, 'Cellist'] = np.nan
    # coord_df.loc[(coord_df['Cellist'].isna()) & (coord_df['Cellist_slice'].notna()), 'Cellist'] = coord_df.loc[(coord_df['Cellist'].isna()) & (coord_df['Cellist_slice'].notna()), 'Cellist_slice']
    q_noise_norm = coord_df_cyto['Cellist_prob_norm'].quantile(noise_prop)
    coord_df['Cellist_norm'] = coord_df['Cellist_unfiltered'].copy()
    coord_df.loc[coord_df['Cellist_prob_norm'] < q_noise_norm, 'Cellist_norm'] = np.nan
    return(coord_df)

def Cellist(platform, resolution, nucleus_seg_method, props_file, nucleus_count_h5_file, nucleus_coord_file, all_spot_count_h5_file, spot_expr_file,
    patch_data_dir, num_workers,
    alpha, sigma, beta, gene_use, max_dist, two_step, cyto, max_dist_s1_scale, noise_prop_s1, noise_prop, neigh_dist, out_dir, out_prefix):
    # ----------- output prefix ----------
    out_prefix = out_prefix + "_Cellist"
    if platform == 'barcoding':
        patch_size = 1100
        stride_length = 100
    elif platform == 'imaging':
        patch_size = 2200
        stride_length = 100
    if two_step:
        out_para = "alpha_%s_sigma_%s_beta_%s_gene_%s_dist_%s_twostep_%s_cyto_%s_noise_%s_neigh_%s_dscale_%s_s1noise_%s" %(alpha, sigma, beta, gene_use, max_dist, two_step, cyto, noise_prop, neigh_dist, max_dist_s1_scale, noise_prop_s1)
    else:
        out_para = "alpha_%s_sigma_%s_beta_%s_gene_%s_dist_%s_twostep_%s_cyto_%s_noise_%s_neigh_%s" %(alpha, sigma, beta, gene_use, max_dist, two_step, cyto, noise_prop, neigh_dist)
    out_dir = os.path.join(out_dir, out_para)
    if not os.path.exists(out_dir):
        os.makedirs(out_dir)
    tmp_dir = os.path.join(out_dir, "tmp")
    if not os.path.exists(tmp_dir):
        os.makedirs(tmp_dir)
    # -------------- write tmp patch files -------------
    if patch_data_dir:
        patch_prefix_list = []
        for patch_file in os.listdir(patch_data_dir):
            if patch_file.endswith('_spot_count.h5'):
                patch_prefix = patch_file.split('_spot_count.h5')[0]
                patch_prefix_list.append(patch_prefix)
    else:
        logging.info('Reading input files...')
        count_df, coord_df, props_df, nucleus_expr_mat, nucleus_gene, nucleus_cell, all_count_mat, all_count_genes, all_count_spots = Read_files(
                nucleus_seg_method, props_file, nucleus_count_h5_file, nucleus_coord_file, all_spot_count_h5_file, spot_expr_file)
        # # ----------- slice patches ----------
        coord_patch_list = slice_coord(coord_df, patch_size, stride_length)
        logging.info('There are %s patches to be processed!' %len(coord_patch_list))
        patch_data_dir = os.path.join(out_dir, "..", "Data_%s" %gene_use)
        if not os.path.exists(patch_data_dir):
            os.mkdir(patch_data_dir)
        res_list = []
        with concurrent.futures.ThreadPoolExecutor(max_workers = 36) as executor:
            for i in range(len(coord_patch_list)):
                logging.info('Writing patch %s...' %i)
                xmin_patch, xmax_patch, ymin_patch, ymax_patch = coord_patch_list[i]
                arg_tuple = (count_df, coord_df, nucleus_seg_method, props_df, nucleus_expr_mat, nucleus_gene, nucleus_cell, all_count_mat, all_count_genes, all_count_spots,
                xmin_patch, xmax_patch, ymin_patch, ymax_patch,
                patch_data_dir, gene_use)
                res_list.append(executor.submit(Write_patch, *arg_tuple))
            done, not_done = concurrent.futures.wait(res_list, timeout=None)
            tmp_input_list_all = [future.result() for future in done]
        del count_df, coord_df, props_df, nucleus_expr_mat, nucleus_gene, nucleus_cell, all_count_mat, all_count_genes, all_count_spots
        gc.collect()
        patch_prefix_list = []
        for patch in tmp_input_list_all:
            if patch != 'NULL':
                patch_prefix_list.append(patch)
    # ----------- segmentation ----------
    logging.info('Start segmentation...')
    patch_res_list = []
    tmp_file_list_all = []
    with concurrent.futures.ProcessPoolExecutor(max_workers = num_workers) as executor_patch:
        for i in range(len(patch_prefix_list)):
            patch_prefix = patch_prefix_list[i]
            logging.info('Processing patch %s...' %patch_prefix)
            props_file_patch = os.path.join(patch_data_dir, "%s_property.txt" %patch_prefix)
            coord_file_patch = os.path.join(patch_data_dir, "%s_coord.txt" %patch_prefix)
            all_spot_count_h5_file_patch = os.path.join(patch_data_dir, "%s_spot_count.h5" %patch_prefix)
            count_file_patch = os.path.join(patch_data_dir, "%s_count.txt" %patch_prefix)
            arg_tuple_patch = (platform, resolution, nucleus_seg_method, props_file_patch, coord_file_patch, all_spot_count_h5_file_patch, count_file_patch,
                out_dir, num_workers, 
                alpha, sigma, beta, gene_use, max_dist, two_step, cyto, max_dist_s1_scale, noise_prop_s1, noise_prop, neigh_dist)
            patch_res_list.append(executor_patch.submit(Soft_assignment_patch, *arg_tuple_patch))
        done_patch, not_done_patch = concurrent.futures.wait(patch_res_list, timeout=None)
        for future in done_patch:
            tmp_file_list_all = tmp_file_list_all + future.result()
    # ------------ merge results from patches ------------
    logging.info('Merging results from patches...')
    count_df, coord_df, _, nucleus_expr_mat, nucleus_gene, nucleus_cell, _, _, _ = Read_files(
            nucleus_seg_method, props_file, nucleus_count_h5_file, nucleus_coord_file, all_spot_count_h5_file, spot_expr_file)
    coord_df_sub_list = []
    for tmp_file in tmp_file_list_all:
        if tmp_file != "NULL":
            coord_df_sub = pd.read_csv(tmp_file, sep = "\t")
            coord_df_sub_list.append(coord_df_sub)
    coord_df_merge = pd.concat(coord_df_sub_list)
    coord_df_merge = coord_df_merge.drop_duplicates(subset = ['x_y', 'Cellist'])
    coord_df_merge = coord_df_merge.dropna(subset = ['Cellist'])
    coord_df_merge = coord_df_merge.sort_values(by=['x_y', 'Cellist_prob'])
    coord_df_merge = coord_df_merge.drop_duplicates(subset = ['x_y'], keep = 'last')
    # coord_df_merge = pd.merge(coord_df, coord_df_merge[['x_y', 'Cellist', 'Cellist_prob', 'Cellist_norm', 'Cellist_prob_norm', 'Cellist_slice']], how = "left", on = "x_y")
    coord_df_merge = pd.merge(coord_df, coord_df_merge[['x_y', 'Cellist', 'Cellist_prob', 'Cellist_norm', 'Cellist_prob_norm']], how = "left", on = "x_y")
    coord_df_merge = Assign_filter(coord_df_merge, out_dir, out_prefix, noise_prop)
    logging.info('Writing segmentation results...')
    coord_df_merge.to_csv(os.path.join(out_dir, "%s_segmentation.txt" %(out_prefix)), sep = "\t", index = False)
    draw_segmentation(coord_df_sub = coord_df_merge, seg_res = "Cellist", out_prefix = out_prefix, 
    out_dir = out_dir, x = "x", y = "y", figsize = (80, 80))
    count_df = pd.merge(count_df, coord_df_merge[['x_y', 'Cellist', 'Nucleus']], how = "left", on = "x_y")
    count_name = count_df.columns[3]
    write_segmentation_h5(count_df, "Cellist", out_prefix, out_dir, count_name)
    write_segmentation_cell_coord(coord_df_merge, "Cellist", out_prefix, out_dir)
    logging.info('Writing parameters and basic cell statistics...')
    hvg_list = get_hvg(nucleus_expr_mat, nucleus_gene, nucleus_cell, 1000)
    cell_num = coord_df_merge['Cellist'].value_counts().shape[0]
    nspot_all = int(coord_df_merge['Cellist'].count())
    nspot_nucleus = int(coord_df_merge[nucleus_seg_method].count())
    nspot_avg = int(nspot_all/cell_num)
    para_dict = {
        'nucleus_seg_method': nucleus_seg_method,
        'props_file': props_file, 
        'nucleus_count_h5_file': nucleus_count_h5_file, 
        'nucleus_coord_file': nucleus_coord_file, 
        'all_spot_count_h5_file': all_spot_count_h5_file, 
        'spot_expr_file': spot_expr_file, 
        'num_workers': num_workers, 
        'alpha': alpha, 
        'sigma': sigma, 
        'beta': beta, 
        'gene_use': gene_use, 
        'max_dist': max_dist, 
        'two_step': two_step,
        'noise_prop': noise_prop,
        'cell_num': cell_num,
        'nspot_nucleus': nspot_nucleus,
        'nspot_all': nspot_all,
        'nspot_avg': nspot_avg
    }
    try:
        if len(nucleus_cell) > 100000:
            seed(0)
            cells_sample = sample(nucleus_cell, 100000)
            Cellist_corr_nucl_cyto, Cellist_corr_nucl_cyto_cell = cal_corr_within_seg_nucleus(count_df.loc[count_df['Cellist'].isin(cells_sample), :], "Cellist", hvg_list, count_name)
        else:
            Cellist_corr_nucl_cyto, Cellist_corr_nucl_cyto_cell = cal_corr_within_seg_nucleus(count_df, "Cellist", hvg_list, count_name)
        Cellist_corr_nucl_cyto_df = pd.DataFrame({'Cell': Cellist_corr_nucl_cyto_cell, 'Correlation': Cellist_corr_nucl_cyto})
        Cellist_corr_nucl_cyto_df.to_csv(os.path.join(out_dir, "%s_corr_nucl_cyto_df.txt" %out_prefix), index = False, sep = "\t")
        # statics
        para_dict['Correlation'] = np.nanquantile(Cellist_corr_nucl_cyto, [0, 0.25, 0.5, 0.75, 1]).tolist()
    except:
        pass
        # write parameters
    with open(os.path.join(out_dir, "parameters.json"), "w") as outfile:
        json.dump(para_dict, outfile)

if __name__ == '__main__':
    parser = CellistParser()
    platform = parser.platform
    resolution = parser.resolution
    spot_expr_file = parser.spot_expr_file
    all_spot_count_h5_file = parser.all_spot_count_h5_file
    nucleus_seg_method = parser.nucleus_seg_method
    props_file = parser.props_file
    nucleus_count_h5_file = parser.nucleus_count_h5_file
    nucleus_coord_file = parser.nucleus_coord_file
    patch_data_dir = parser.patch_data_dir
    num_workers = parser.num_workers
    max_dist = parser.max_dist
    neigh_dist = parser.neigh_dist
    noise_prop = parser.noise_prop
    alpha = parser.alpha
    sigma = parser.sigma
    beta = parser.beta
    gene_use = parser.gene_use
    two_step = parser.two_step
    cyto = parser.cyto
    out_dir = parser.out_dir
    out_prefix = parser.out_prefix
    max_dist_s1_scale = parser.max_dist_s1_scale
    noise_prop_s1 = parser.noise_prop_s1
    Cellist(platform, resolution, nucleus_seg_method, props_file, nucleus_count_h5_file, nucleus_coord_file, all_spot_count_h5_file, spot_expr_file,
        patch_data_dir, num_workers, alpha, sigma, beta, gene_use, max_dist, two_step, cyto, max_dist_s1_scale, noise_prop_s1, noise_prop, neigh_dist,
        out_dir, out_prefix)



# -*- coding: utf-8 -*-
# @Author: dongqing
# @Date:   2025-04-17 13:26:55
# @Last Modified by:   dongqing
# @Last Modified time: 2025-04-17 14:15:49


import gc
import json
import os, sys
import argparse
import functools
import pandas as pd
import numpy as np
import multiprocessing as mp
import seaborn as sns
import skimage
from scipy.spatial import KDTree, distance
from sklearn.preprocessing import StandardScaler
import concurrent.futures
from random import sample, seed
from anndata import ImplicitModificationWarning

from Cellist.Utility import *
from Cellist.IO import *
from Cellist.Plot import *
from Cellist.Segmentation import Read_files, Assign_filter

import logging
logging.basicConfig(level=logging.INFO, format='%(asctime)s  %(message)s')

import warnings
pd.set_option('mode.chained_assignment', None)
warnings.filterwarnings('ignore', category=RuntimeWarning)
warnings.filterwarnings('ignore', category=ImplicitModificationWarning, append=True)

def TuneParser():
    parser = argparse.ArgumentParser(
                        description='Tune the --noise_prop based on the segmentation results with other parameters fixed. ')
    group_input = parser.add_argument_group("Input arguments")
    group_input.add_argument("--gem", dest = "spot_expr_file", default = None,
        help = "Bin1 gem file generated by SAW pipeline. ")
    group_input.add_argument("--spot-count-h5", dest = "all_spot_count_h5_file", default = None,
        help = "Bin1 spot-level expression file in h5 format.")
    group_input.add_argument("--nucleus-seg-method", dest = "nucleus_seg_method", default = "Watershed", choices = ['Watershed', 'Cellpose'],
        help = "The initial method to segment nuclei, Watershed or Cellpose. ")
    group_input.add_argument("--nucleus-prop", dest = "props_file", default = None,
        help = "Properties of segmented nucleus by Watershed or Cellpose algorithm. ")
    group_input.add_argument("--nucleus-count-h5", dest = "nucleus_count_h5_file", default = None,
        help = "Nuclei-level expression generated by Watershed or Cellpose segmentation with h5 format. ")
    group_input.add_argument("--nucleus-seg", dest = "nucleus_coord_file", default = None,
        help = "Spot-level Watershed or Cellpose segmentation result. ")
    group_input.add_argument("--cellist-seg", dest = "cellist_coord_file", default = None,
        help = "Existing spot-level Cellist segmentation result. ")

    group_model = parser.add_argument_group("Model arguments")
    group_model.add_argument("--cell-radius", dest = "max_dist", type = int, default = 15,
        help = "Expected cell radius (in microns). "
        "Considering the diffusion of transcripts, it should be 1.5X~2X larger than the real cell size. "
        "Default: 15")
    group_model.add_argument("--spot-imputation-distance", dest = "neigh_dist", type = float, default = 2.5,
        help = "Imputation distance for each spot (in microns). Default: 2.5. ")
    group_model.add_argument("--noise-prop", dest = "noise_prop", type = float, default = 0.25,
        help = "Proportion of noise spots. Default: 0.25. ")
    group_model.add_argument("--alpha", dest = "alpha", type = float, default = 0.8,
        help = "Weight for expression distance. Default: 0.8. ")
    group_model.add_argument("--sigma", dest = "sigma", type = float, default = 1.0,
        help = "Weight for regularization. Default: 1.0. ")
    group_model.add_argument("--beta", dest = "beta", type = float, default = 10,
        help = "Penalty weight. Default: 10. ")
    group_model.add_argument("--gene-use", dest = "gene_use", type = str, default = 'HVG', choices = ['HVG', 'Frequent'],
        help = "Gene to use to calculate spot-cell distance. "
        "'HVG' means highly variable genes, and 'Frequent' means top frequent genes. "
        "Default: HVG. ")
    group_model.add_argument("--two-step", dest = "two_step", action = "store_true",
        help = "Two-step mode. If set, Cellist will assign non-nuclei spots recusively. ")
    group_model.add_argument("--cyto", dest = "cyto", action = "store_true", 
        help = "Two-step cyto mode. If set, Cellist will calculate the correlation between non-nuclei spots and the 1st cycle of cytoplasm. ")

    group_output = parser.add_argument_group("Output arguments")
    group_output.add_argument("--outdir", dest = "out_dir", default = "",help = "Output directory.")
    group_output.add_argument("--outprefix", dest = "out_prefix", default = "",help = "Output prefix.")
    return(parser)

def Tune(nucleus_seg_method, props_file, nucleus_count_h5_file, nucleus_coord_file, cellist_coord_file, all_spot_count_h5_file, spot_expr_file,
    alpha, sigma, beta, gene_use, max_dist, two_step, cyto, noise_prop, neigh_dist, out_dir, out_prefix):
    # ----------- output prefix ----------
    out_para = "alpha_%s_sigma_%s_beta_%s_gene_%s_dist_%s_twostep_%s_cyto_%s_noise_%s_neigh_%s" %(alpha, sigma, beta, gene_use, max_dist, two_step, cyto, noise_prop, neigh_dist)
    out_dir = os.path.join(out_dir, out_para)
    if not os.path.exists(out_dir):
        os.makedirs(out_dir)
    # ------------ merge results from patches ------------
    logging.info('Merging results from patches...')
    count_df, coord_df, _, nucleus_expr_mat, nucleus_gene, nucleus_cell, _, _, _ = Read_files(
            nucleus_seg_method, props_file, nucleus_count_h5_file, nucleus_coord_file, all_spot_count_h5_file, spot_expr_file)
    coord_df_merge = pd.read_csv(cellist_coord_file , sep = "\t")
    coord_df_merge['Cellist'] = coord_df_merge['Cellist_unfiltered'].copy()
    coord_df_merge = Assign_filter(coord_df_merge, out_dir, out_prefix, noise_prop)
    logging.info('Writing segmentation results...')
    coord_df_merge.to_csv(os.path.join(out_dir, "%s_segmentation.txt" %(out_prefix)), sep = "\t", index = False)
    draw_segmentation(coord_df_sub = coord_df_merge, seg_res = "Cellist", out_prefix = out_prefix, 
    out_dir = out_dir, x = "x", y = "y", figsize = (80, 80))
    count_df = pd.merge(count_df, coord_df_merge[['x_y', 'Cellist', 'Nucleus']], how = "left", on = "x_y")
    count_name = count_df.columns[3]
    write_segmentation_h5(count_df, "Cellist", out_prefix, out_dir, count_name)
    write_segmentation_cell_coord(coord_df_merge, "Cellist", out_prefix, out_dir)
    logging.info('Writing parameters and basic cell statistics...')
    hvg_list = get_hvg(nucleus_expr_mat, nucleus_gene, nucleus_cell, 1000)
    cell_num = coord_df_merge['Cellist'].value_counts().shape[0]
    nspot_all = int(coord_df_merge['Cellist'].count())
    nspot_nucleus = int(coord_df_merge[nucleus_seg_method].count())
    nspot_avg = int(nspot_all/cell_num)
    para_dict = {
        'nucleus_seg_method': nucleus_seg_method,
        'props_file': props_file, 
        'nucleus_count_h5_file': nucleus_count_h5_file, 
        'nucleus_coord_file': nucleus_coord_file, 
        'all_spot_count_h5_file': all_spot_count_h5_file, 
        'spot_expr_file': spot_expr_file, 
        'alpha': alpha, 
        'sigma': sigma, 
        'beta': beta, 
        'gene_use': gene_use, 
        'max_dist': max_dist, 
        'two_step': two_step,
        'noise_prop': noise_prop,
        'cell_num': cell_num,
        'nspot_nucleus': nspot_nucleus,
        'nspot_all': nspot_all,
        'nspot_avg': nspot_avg
    }
    try:
        if len(nucleus_cell) > 100000:
            seed(0)
            cells_sample = sample(nucleus_cell, 100000)
            Cellist_corr_nucl_cyto, Cellist_corr_nucl_cyto_cell = cal_corr_within_seg_nucleus(count_df.loc[count_df['Cellist'].isin(cells_sample), :], "Cellist", hvg_list, count_name)
        else:
            Cellist_corr_nucl_cyto, Cellist_corr_nucl_cyto_cell = cal_corr_within_seg_nucleus(count_df, "Cellist", hvg_list, count_name)
        Cellist_corr_nucl_cyto_df = pd.DataFrame({'Cell': Cellist_corr_nucl_cyto_cell, 'Correlation': Cellist_corr_nucl_cyto})
        Cellist_corr_nucl_cyto_df.to_csv(os.path.join(out_dir, "%s_cellist_corr_nucl_cyto_df.txt" %out_prefix), index = False, sep = "\t")
        # statics
        para_dict['Correlation'] = np.nanquantile(Cellist_corr_nucl_cyto, [0, 0.25, 0.5, 0.75, 1]).tolist()
    except:
        pass
        # write parameters
    with open(os.path.join(out_dir, "parameters.json"), "w") as outfile:
        json.dump(para_dict, outfile)

if __name__ == '__main__':
    parser = TuneParser().parse_args()
    spot_expr_file = parser.spot_expr_file
    all_spot_count_h5_file = parser.all_spot_count_h5_file
    nucleus_seg_method = parser.nucleus_seg_method
    props_file = parser.props_file
    nucleus_count_h5_file = parser.nucleus_count_h5_file
    nucleus_coord_file = parser.nucleus_coord_file
    cellist_coord_file = parser.cellist_coord_file
    max_dist = parser.max_dist
    neigh_dist = parser.neigh_dist
    noise_prop = parser.noise_prop
    alpha = parser.alpha
    sigma = parser.sigma
    beta = parser.beta
    gene_use = parser.gene_use
    two_step = parser.two_step
    cyto = parser.cyto
    out_dir = parser.out_dir
    out_prefix = parser.out_prefix
    Tune(nucleus_seg_method, props_file, nucleus_count_h5_file, nucleus_coord_file, cellist_coord_file, all_spot_count_h5_file, spot_expr_file,
        alpha, sigma, beta, gene_use, max_dist, two_step, cyto, noise_prop, neigh_dist,
        out_dir, out_prefix)




# -*- coding: utf-8 -*-
# @Author: dongqing
# @Date:   2025-03-08 16:01:35
# @Last Modified by:   Dongqing
# @Last Modified time: 2025-08-26 13:15:17


import sys, os, gc
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns

import torch
import matplotlib
matplotlib.rcParams['font.family'] = 'sans-serif'
matplotlib.rcParams['font.sans-serif'] = ['Arial']

import logging
logging.basicConfig(level=logging.INFO, format='%(asctime)s  %(message)s')

from cellpose import core, utils, io, models, metrics

from skimage.io import imread, imsave
from skimage.filters import threshold_otsu, threshold_multiotsu, threshold_niblack, threshold_local
from skimage.feature import peak_local_max
from skimage.measure import label, regionprops_table
from skimage.segmentation import watershed, expand_labels, mark_boundaries
from skimage.color import label2rgb, rgb2gray, rgb2hed
from skimage.segmentation import mark_boundaries
from skimage.morphology import remove_small_holes

from scipy.sparse import csc_matrix, save_npz, load_npz, csr_matrix
from scipy import ndimage as ndi

from Cellist.Utility import *
from Cellist.IO import gem_to_mat
from Cellist.Plot import *
from Cellist.Watershed import Thresholding, write_segmentation_coord, write_segmentation_cell

use_GPU = core.use_gpu()
print('>>> GPU activated? %s'%use_GPU)

'''
cellist watershed --gem /mnt/Venus/home/dongxin/Projects/heart/202401_Stereo_seq/result_C02942B3/041.cellcut/C02942B3.adjusted.gem \
--tif Alignment/ssDNA_C02942B3_regist_regist_transposed_aligned_by_Spateo.tiff \
--outdir Watershed --outprefix C02942B3

cellist watershed --gem /fs/home/dongxin/Projects/heart/202401_Stereo_seq/result_C02942B3/041.cellcut/C02942B3.adjusted.gem \
--tif Alignment/ssDNA_C02942B3_regist_regist_transposed_aligned_by_Spateo.tiff \
--outdir Watershed --outprefix C02942B3
'''

def CellposeParser(subparsers):
    parser = subparsers.add_parser("cellpose", 
        help = "Run initial cellpose segmentation on the staining image. ")
    group_input = parser.add_argument_group("Input arguments")
    group_input.add_argument("--platform", dest = "platform", default = 'barcoding', choices = ['barcoding', 'imaging'],
        help = "Supported high-resoultion ST platform: 'barcoding' (e.g., Stereo-seq, Seq-scope) or "
        "'imaging' (e.g., seqFISH, STARmap, 10x Xenium). Default: barcoding.")
    group_input.add_argument("--gem", dest = "spot_expr_file", default = None,
        help = "Bin1 gem file generated by SAW pipeline located in '041.cellcut' or '04.tissuecut'. ")
    group_input.add_argument("--tif", dest = "regist_tif_file", default = None,
        help = "Registered tif file generated by SAW pipeline located in '04.tissuecut' or the refined tif file generated by 'cellist align'. ")
    group_input.add_argument("--no-local-threshold", dest = "no_local_threshold", action = "store_true", 
        help = "Whether or not to perform local thresholding on the image. "
        "If not set, Cellist will perform local thresholding. If set, the local thresholding will be skipped. ")
    group_input.add_argument("--diameter", dest = "diameter", type = int, default = 15,
        help = "The average cell/nucleus diameter (in pixel) to run cellpose model. Default: 15. ")
    group_input.add_argument("--model-type", dest = "model_type", default = 'nuclei', choices = ["nuclei", "cyto", "cyto2", "cyto3"],
        help = "Cellpose model type. cyto: cytoplasm model; nuclei: nucleus model; "
        "cyto2: cytoplasm model with additional user images; "
        "cyto3: super-generalist model. Default: nuclei. ")
    group_input.add_argument("--flow-threshold", dest = "flow_threshold", type = float, default = 0.4,
        help = "The flow_threshold parameter in Cellpose is the maximum allowed error of the flows for each mask. "
        "Increase this threshold if cellpose is not returning as many ROIs as you’d expect. "
        "Similarly, decrease this threshold if cellpose is returning too many ill-shaped ROIs. Default: 0.4.")
    group_input.add_argument("--cellprob-threshold", dest = "cellprob_threshold", type = float, default = 0.0,
        help = "The cellprob_threshold parameter in Cellpose, varies from around -6 to +6. "
        "Decrease this threshold if cellpose is not returning as many ROIs as you’d expect. "
        "Similarly, increase this threshold if cellpose is returning too ROIs particularly from dim areas. Default: 0.0")
    group_input.add_argument("--expansion", dest = "expansion", action = "store_true", 
        help = "Whether or not to perform expansion after Cellpose segmentation. "
        "If set, expansion based on Cellpose segmentation will be performed. ")
    group_input.add_argument("--expansion-dist", dest = "expansion_dist", type = int, default = 8,
        help = "The expansion distance (in pixel). Default: 8. ")

    group_output = parser.add_argument_group("Output arguments")
    group_output.add_argument("--outdir", dest = "out_dir", default = '.',
        help = "The output directory. ")
    group_output.add_argument("--outprefix", dest = "out_prefix", default = None,
        help = "The output prefix. ")
    return(parser)

def cellpose_seg(img, otsu_local_hole_rm, out_dir, out_img_prefix, diameter = 15, model_type = 'nuclei', flow_threshold = None, 
    cellprob_threshold = 0, expansion = False, expansion_dist = 8):
    logging.info("Cellpose segmentation... ")
    # 0. load model
    model = models.Cellpose(gpu=use_GPU, model_type = model_type, device = torch.device('cuda:1'))
    channels = [0, 0]
    segmented_cells, flow, style, diam = model.eval(otsu_local_hole_rm, diameter = diameter, flow_threshold = flow_threshold, channels=channels, cellprob_threshold = cellprob_threshold)
    logging.info("Number of segments: %s" %segmented_cells.max())
    cellpose_seg_file = os.path.join(out_dir, "%s_Cellpose_nucleus_matrix.npz" %out_img_prefix)
    segmented_cells_mat = csc_matrix(segmented_cells, dtype=np.int32)
    save_npz(cellpose_seg_file, segmented_cells_mat)
    fig, ax = plt.subplots(figsize=(20, 20))
    ax.imshow(mark_boundaries(img, segmented_cells, color = (1,0,0)))
    plt.savefig(os.path.join(out_dir, "%s_Cellpose_boundary.png" %out_img_prefix), bbox_inches='tight')
    props = regionprops_table(segmented_cells, 
                              properties=['label', 'area','centroid',
                                          'equivalent_diameter_area'])
    props_df = pd.DataFrame(props)
    props_df.to_csv(os.path.join(out_dir, "%s_Cellpose_nucleus_property.txt" %out_img_prefix), sep = "\t", index = False)
    if expansion:
        expanded = expand_labels(segmented_cells, distance = expansion_dist)
        cellpose_seg_file = os.path.join(out_dir, "%s_Cellpose_cells_expansion%s_matrix.npz" %(out_img_prefix, expansion_dist))
        expanded_mat = csc_matrix(expanded, dtype=np.int32)
        save_npz(cellpose_seg_file, expanded_mat)
        props_expanded = regionprops_table(expanded, 
                                  properties=['label', 'area','centroid',
                                              'equivalent_diameter_area'])
        props_df_expanded = pd.DataFrame(props_expanded)
        props_df_expanded.to_csv(os.path.join(out_dir, "%s_Cellpose_cells_expansion%s_property.txt" %(out_img_prefix, expansion_dist)), sep = "\t", index = False)
        return segmented_cells, expanded
    else:
        return segmented_cells

def Cellpose(platform, gem_path, img_path, out_dir, out_prefix, no_local_threshold, diameter, model_type, flow_threshold, cellprob_threshold, expansion, expansion_dist):
    if not out_prefix:
        file_name, file_extension = os.path.splitext(gem_path)
        base_name = os.path.basename(file_name)
        out_gem_prefix = base_name
        file_name, file_extension = os.path.splitext(img_path)
        base_name = os.path.basename(file_name)
        out_img_prefix = base_name
    else:
        out_gem_prefix = out_prefix
        out_img_prefix = out_prefix
    if not os.path.exists(out_dir):
        os.makedirs(out_dir)
    # --------------------- read image file and thresholding --------------------- 
    threshold_file = os.path.join(out_dir, "%s_thresholding.tif" %out_img_prefix)
    if os.path.exists(threshold_file):
        otsu_local_hole_rm = imread(threshold_file)
        img = imread(img_path)
    else:
        img, otsu_local_hole_rm = Thresholding(platform, img_path, out_dir, out_img_prefix, no_local_threshold)
    # --------------------- watershed segmentation --------------------- 
    if expansion:
        segmented_cells, expanded = cellpose_seg(img, otsu_local_hole_rm, out_dir, out_img_prefix, diameter = diameter, model_type = model_type,
        flow_threshold = flow_threshold, cellprob_threshold = cellprob_threshold,
        expansion = expansion, expansion_dist = expansion_dist)
    else:
        segmented_cells = cellpose_seg(img, otsu_local_hole_rm, out_dir, out_img_prefix, diameter = diameter, model_type = model_type,
        flow_threshold = flow_threshold, cellprob_threshold = cellprob_threshold,
        expansion = expansion, expansion_dist = expansion_dist)
    # --------------------- read gem file --------------------- 
    logging.info("Reading count file...")
    gem_df = pd.read_csv(gem_path, sep = "\t", comment = "#")
    bin1_out_file = os.path.join(out_dir, "%s_bin1.h5" %out_gem_prefix)
    countname = gem_df.columns[3]
    gem_df = gem_to_mat(gem_df = gem_df, outfile = bin1_out_file, countname = countname)
    coord_df = gem_df.drop_duplicates(subset = ['x_y'])
    coord_df = coord_df[["x", "y", "x_y"]]
    # --------------------- write segmentation result --------------------- 
    if expansion:
        nucleus_barcode_df, expanded_barcode_df = write_segmentation_coord(segmented_cells, coord_df, out_dir, out_gem_prefix, seg_method = "Cellpose", expansion = expansion, expanded = expanded, expansion_dist = expansion_dist)
    else:
        nucleus_barcode_df = write_segmentation_coord(segmented_cells, coord_df, out_dir, out_gem_prefix, seg_method = "Cellpose")
        expanded_barcode_df = None
    write_segmentation_cell(nucleus_barcode_df, gem_df, out_gem_prefix, out_dir, seg_method = "Cellpose", expansion = expansion, expanded_barcode_df = expanded_barcode_df, expansion_dist = expansion_dist)
    logging.info("Drawing segmentation plot...")
    draw_segmentation(nucleus_barcode_df, seg_res = "Cellpose", out_prefix = "%s_Cellpose" %out_gem_prefix,
        out_dir = out_dir, x = "x", y = "y", figsize = (80, 80))
    if expansion:
        draw_segmentation(expanded_barcode_df, seg_res = 'Cellpose_expansion_%s' %(expansion_dist), out_prefix = "%s_Cellpose_expansion_%s" %(out_gem_prefix, expansion_dist),
            out_dir = out_dir, x = "x", y = "y", figsize = (80, 80))
    logging.info("Drawing segmentation plot is done.")
    logging.info("All done!")

if __name__ == '__main__':
    parser = WatershedParser()
    platform = parser.platform
    gem_path = parser.spot_expr_file
    img_path = parser.regist_tif_file
    out_dir = parser.out_dir
    out_prefix = parser.out_prefix
    no_local_threshold = parser.no_local_threshold
    diameter = parser.diameter
    model_type = parser.model_type
    flow_threshold = parser.flow_threshold
    cellprob_threshold = parser.cellprob_threshold
    expansion = parser.expansion
    expansion_dist = parser.expansion_dist
    Cellpose(platform = platform, gem_path = gem_path, img_path = img_path, out_dir = out_dir, out_prefix = out_prefix, no_local_threshold = no_local_threshold,
         diameter = diameter, model_type = model_type,flow_threshold = flow_threshold, cellprob_threshold = cellprob_threshold, expansion = expansion, expansion_dist = expansion_dist)
